/* Dean.c
 * Jim Gochee
 * 4/2/92
 *
 * This file is the main source file for the 'dean', which is a daemon that
 * runs on each machine that is participating in a SPEDE environment. The
 * dean is the middleman between objects.
 *
 */
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <varargs.h>
#include <utmp.h>
#include <sys/file.h>

#include "sockets.h"
#include "select.h"
#include "network.h"
#include "spede.h"
#include "timed_task.h"
#include "objects.h"
#include "misc.h"
#include "spede_io.h"
#include "dean_objs.h"
#include "dean.h"

#define TRUE 1
#define FALSE 0
#define DIE_MINUTES  10

/*#define DEBUG */

extern int errno;
extern char *optarg;
extern int optind, opterr;

/*****************************************************************************
 *                                                                           *
 *                               Global Variables                            *
 *                                                                           *
 *****************************************************************************/

connect_type *connections;
int world_connect;                   /* The main socket to the world         */
int world_port;
long my_addr;              /* This node's address                  */
FILE *debug_file;                    /* fileno used for the debug file       */
node_des local_node;                 /* This host's info                     */
char *program = "dean";              /* Name of this program                 */
int debug_on = FALSE;
long last_io_time;
char local_hostname[33];
int connected_serv, service_socket;
char *base_dir;
char reghost[256];
int reg_port;

/*****************************************************************************
 *                                                                           *
 *                          Utility Functions                                *
 *                                                                           *
 *****************************************************************************/

/* LOG
 *
 * This routine writes output to the debug_file.
 */
void log(format, va_alist)
     char *format;
     va_dcl
{
  va_list ap;
  
  if (!debug_on)
    return;

  va_start(ap);
  vfprintf(debug_file, format, ap);
  fflush(debug_file);
  va_end(ap);
}


/*****************************************************************************
 *                                                                           *
 *                            Functions                                      *
 *                                                                           *
 *****************************************************************************/


void kill_dean(comm)
     command *comm;
{
  log("DEAN - SHUTDOWN...asked to shutdown on socket %d\n", comm->socket);
  exit(0);
}


/* DEAN_COMMAND
 *
 * This routine determines which command was received and acts on it
 * accordinly.
 */
int dean_command(comm, ce)
     command *comm;
     connect_entry *ce;
{
  int error;
  int the_socket = ce->the_socket;
  double d;
  short i;
  int err;
  
  log("DEAN - command %s from socket %d\n", command_name[comm->comm], ce->the_socket);

  switch(comm->comm) {
  case REQUEST_ID:
    comm->data.req_id.connect_id = ((connect_des *) ce->data)->connect_id;
    write_command(comm, ce->the_socket);
    break;
  case NEW_OBJECT:
    f_new_object(comm);
    break;
  case OBJECT_CREATED:
    f_object_created(comm, ce);
    break;
  case DELETE_OBJECT:
    f_delete_object(comm);
    break;
  case OBJECT_DELETED:
    f_object_deleted(comm);
    break;
  case OBJECT_LOAD:
    f_object_load(comm, ce);
    break;
  case OBJECT_MESSAGE:
    f_object_message(comm);
    break;
  case RETURN_MESSAGE:
    f_return_message(comm);
    break;
  case LINK_OBJECT:
    f_link_object(comm);
    break;
  case SPEED_LINK:
    f_speed_link(comm);
    break;
  case KILL:
    kill_dean(comm);
    break;
  case START_TIME:
    f_start_time(comm);
    break;
  case GET_TIME:
    f_get_time(comm);
    break;
  case SEND_TIMES:
    f_send_times(comm);
    break;
  default:
    log("DEAN - Unknown command #%d\n", comm->comm);
    break;
  }

  return(0);
}


/* SOCKET_WENT_DOWN
 *
 * This routine is called by the main_loop when a socket has an error (i.e. it
 * lost its connection).
 */
int socket_went_down(ce)
     connect_entry *ce;
{
  close_commands(ce->the_socket);

  if (ce->the_socket == service_socket)
    connected_serv = FALSE;
  else {
    delete_all_objects(ce->the_socket);   /* Delete objects using the socket */
    log("DEAN - closing socket %d\n", ce->the_socket);
    free(ce->data);
  }
  delete_connection(connections, ce);
}


int f_check_time(data)
     void *data;
{
  if (time((time_t *) 0) - last_io_time > DIE_MINUTES*60)
    return(-1);
  else
    return(0);
}

int f_check_serv(data)
     void *data;
{
  if (!connected_serv)
    connect_service();
  return(0);
}


int connect_service()
{
  int socket;
  command comm;
  struct hostent *hp;

  if (connected_serv)
    return;

  gethostname(local_hostname, sizeof(local_hostname));

  socket = connect_register_serv();
  if (socket == -1) {
    fprintf(stderr, "Dean is quitting. Could not connect with register\n");
    connected_serv = FALSE;
    return(-1);
  }

  register_service(local_hostname, "dean", world_port, &base_dir);
  connected_serv = TRUE;
  service_socket = socket;
  add_connection(connections, socket, &local_node);
  open_commands(socket);
  return(socket);
}

  


/* MAIN_LOOP
 *
 * Main_loop loops checking for commands and new connections. If a command
 * is found, a command handler is called. If a new connection is forming, a
 * call to accept it is made. Socket errors are caught in the command routine.
 */
int main_loop()
{
  int err;
  int status;
  struct timeval ti, notify_time;
  struct timezone tz;
  int nfound, new_socket, socket, test_connect, result;
  connect_entry *ce;
  long output_time, check_time, check_serv;
  long check_id;
  timedtasks ts;
  int new_sock;
  command comm;
  
  last_io_time = time((time_t *) 0);
  connections = new_connections();
  ce = add_connection(connections, world_connect, &local_node);

  ts = new_timed_tasks();
  check_time = add_timed_task(ts, 30, 0, f_check_time, (void *) 0);
  check_serv = add_timed_task(ts, 10, 0, f_check_serv, (void *) 0);

  connect_service();

  while(TRUE) {
    next_task_wait(ts, &ti); 
    nfound = wait_on_connections(connections, &ti);    

    /* Loop for the number of sockets with data */
    if (nfound) {
      last_io_time = time((time_t *) 0);
      while(ce = get_next_entry(connections)) {
	if (ce->the_socket == world_connect) {
	  ce = accept_new_connection(connections, world_connect);
	  open_commands(ce->the_socket);
	  new_connection(ce, OBJECT);
	  log("DEAN - Accepted new connection on socket %d\n", ce->the_socket);
	}
	else {                                 /* Must be a command */
	  log("DEAN - data on socket %d\n", ce->the_socket);
	  if ((status = complete_command(&comm, ce->the_socket))) {
	    if (status == -1)
	      socket_went_down(ce);
	    else
	      dean_command(&comm, ce);
	  }
	}
      }
    }
    result = do_tasks(ts);
    if (result) {
      log("DEAN - No user interaction in a while... exiting\n");
      break;
    }
  }
  
  free_timed_tasks(ts);
  return(result);
}


/* USAGE
 *
 * Describe the commands possible with 'dean'
 */
void usage()
{
  fprintf(stderr, "usage: %s [options]\n", program);
  fprintf(stderr, "  where options are:\n");
  fprintf(stderr, "  -h <hostname>        ; REQUIRED name of reg host\n");
  fprintf(stderr, "  -p <port number>     ; REQUIRED port number of reg\n");
  fprintf(stderr, "  -d                   ; print debug statements\n");
  fprintf(stderr, "\nExample:\n");
  fprintf(stderr, "  dean -h everest -p 7000\n");
}


/* SET_OPTIONS
 *
 * This routine sets command line options.
 *
 * The only option allowed is -d, which prints debuging information to
 * stdout.
 */
void set_options(argc, argv)
     int argc;
     char **argv;
{
  int ch;
  int go_stdout = FALSE;
  
  while((ch = getopt(argc, argv, "h:p:d?")) != EOF)
    switch(ch) {
    case 'd':
      debug_on = TRUE;
      break;
    case '?':
      break;
    case 'h':
      strcpy(reghost, optarg);
      break;
    case 'p':
      reg_port = atoi(optarg);
      break;
    default:
      usage();
      exit(1);
    }
  

  if (reg_port == 0 || reghost[0] == '\0') {
    usage();
    exit(0);
  }

  register_info(reghost, reg_port);

  debug_file = stdout; 

  fclose(stdin);
  fclose(stderr);
  if (!debug_on)
    fclose(stdout);
}


main(argc, argv)
     int argc;
     char **argv;
{
  int err;
  char fname[256];
  
/*  if (fork())
    exit(0); */

  set_options(argc, argv);
  random_listener(&world_connect, &world_port, 20);

  if (world_connect == -1) {
    printf("Error trying to install a socket listener.");
    exit(1);
  }
  
  get_local_address(&my_addr);
  my_getnodebyaddr(my_addr, &local_node);
  
  log("\nDEAN - Local machine's address: %X\n", ntohl(my_addr)); 
  
  init_objects();
  main_loop();
}
