/* SPEDE_IO.H
 * Jim Gochee
 * 5/14/92
 *
 */

/***************************************************************************
 *                                                                         *
                            Network Protocol Info
 *                                                                         *
 ***************************************************************************/


#define PROTO_ERROR     0x00
#define OK              0x01
#define NEW_OBJECT      0x02  /* Sent by obj to dean to create an object */
#define CREATION_FAILED 0x03  /* Sent by dean to obj when creation fails */
#define OBJECT_CREATED  0x04  /* Sent by dean to obj when creation works */
#define OBJECT_LOAD     0x05  /* Sent by object to dean when object starts */
#define OBJECT_MESSAGE  0x06  /* Sent by obj to obj to do work */
#define RETURN_MESSAGE  0x07  /* Sent as a return from OBJECT_MESSAGE */
#define DELETE_OBJECT   0x08  /* Sent to obj when obj is being deleted */
#define OBJECT_DELETED  0x09  /* Sent by dean to calling object */
#define REQUEST_ID      0x0A  /* Sent by obj to dean to get a unique id */
#define LOG_MESSAGE     0x0B  /* Sent by dean to obj to log a mess/ret pair */
#define RETURN_STATE    0xOC  /* Sent by obj to dean to get object state */
#define STATE_INFO      0x0D  /* Sent by dean to obj on state info request */
#define OBJECT_CRASHED  0x0E  /* Sent by dean to obj when object crashes */
#define SET_OBJ_STATE   0x0F  /* Used for migration or restarting an object */
#define FREE_LAUNCH     0x10  /* NOT USED */
#define LINK_OBJECT     0x11  /* Sent to dean when a link is being made */
#define OBJECT_LOADED   0x12  /* Sent to obj when a free launch happens */
#define OBJECT_LINKED   0x13  /* Sent to obj when a link was successful */
#define SPEED_LINK      0x14  /* Sent to dean to form a speed link */
#define SPEED_MADE      0x15  /* Sent to obj when a speed link is successful */
#define KILL            0x16  /* Sent by REG to stop a dean */
#define REGISTER_SERV   0x17  /* Sent by obj to REG to notify of existence */
#define NO_OP           0x18  /* Sent by REG to dean as a tickle */
#define LAUNCH_REMOTE   0x19  /* NOT USED */
#define GET_CLASS_INFO  0x1A  /* Sent by obj to REG to get class information */
#define R_CLASS_INFO    0x1B  /* Sent by REG to obj with class information */
#define LAUNCH_SERVICE  0x1C  /* Sent by obj to REG to launch a dean */
#define START_TIME      0x1D  /* Resets internal obj timing info */
#define GET_TIME        0x1E  /* Gets internal obj timing info */
#define SEND_TIMES      0x1F  /* Sending internal obj timing info */
#define QUERY_SERV      0x20  /* Sent to REG from obj to get dean port num */
#define SERV_INFO       0x21  /* Sent to obj from REG with dean port num */
#define BASE_DIR        0x22  /* Sent to obj from REG with obj base dir */


/***************************************************************************
 *                                                                         *
                                TypeDefs
 *                                                                         *
 ***************************************************************************/


/* The command type is the backbone of the network protocol. Every message
 * that is sent between two SPEDE entities must be in the command format.
 * The task of reading and writing commands is removed to one library file
 * (spede_io.c) so that no other applications need to know anything about the
 * low level transport.
 */

typedef struct _command {
  long num_bytes;                        /* Number of bytes coming */
  char comm;                             /* The command being sent */
  long message_id;                       /* ID of the message */  
  long time_sent;                        /* A relative timestamp */
  long recv_ref;                         /* The receiver's reference num */
  long send_ref;                         /* The sender's reference num */
  union {
    struct {
      char type[30];
      char machine[33];
      long port_num;
    } reg_serv;
    struct {
      unsigned char connect_id;
    } req_id;
    struct {
      object_instance object;
    } new_o;
    struct {
      short error_code;
    } n_created;
    struct {
      long proc_id;
      char object_name[32];
      long port_num;
    } obj_load;
    struct {
      short status;                        /* 0 = free launch, 1 = loop */
    } done_load;
    struct {
      short the_message;
      parameter_block *params;
      short wait_return;
    } mess;
    struct {
      long result;
      parameter_block *params;
    } return_m;
    struct {
      struct _command *comm;
    } log;
    struct {
      short status;                      /* 0 = successful link, 1 = failed */
      object_instance object;
    } o_linked;
    struct {
      short status;
      long port_num;                     /* The object's port num */
    } speed_made;
    struct {
      short control;                       /* 0=get info, 1=get machines */
      short num;                           /* number of machines to get */
      char class[33];                      /* the class in question */
    } class;
    struct {
      short num_machines;
      char **machine;                     /* a list of machines */
    } r_class;
    struct {                              /* Launch a service */
      char machine[33];
      char service[33];
    } l_serv;
    struct {
      double time_idle;
      double time_computing;
      double time_writing;
    } send_times;
    struct {
      char machine[33];
      char service[33];
    } query_serv;
    struct {
      char machine[33];
      char service[33];
      long port_num;
      short running;
    } serv_info;
    struct {
      char base[100];
    } base_dir;
  } data;
  
  int socket;                           /* The socket the command came from */
} command;

#define OBJECT_ID(a)  ((a)->message_id & 0x00FFFFFF)
#define LOCAL_ID(a)   ((a)->message_id & 0x0000FFFF)
#define MESSAGE_ID(a) (((a)->message_id & 0xFF000000) >> 24)
#define CONNECT_ID(a) (((a)->message_id & 0x00FF0000) >> 16)

extern char *command_name[];

/** LOWER LEVEL I/O ROUTINES **/

#ifdef __STDC__
void write_params(parameter_block *);
void write_object(object_instance *);
void write_node_des(node_des *);
void write_link_des(link_des *);
int read_object(int, object_instance *);
int read_node_des(int, node_des *);
int read_link_des(int, link_des *);
int read_params(int, parameter_block **);


/** COMMAND ROUTINES **/

command *new_command(void);
int read_command(command *c, int socket);
int write_command(command *c, int socket);
void clean_command(command *c);
void free_command(command *c);
int complete_command(command *c, int socket);
int complete_a_command(command **c, int socket);
void open_commands(int socket);
void close_commands(int socket);


/** COMMAND CONSTRUCTION ROUTINES **/

void build_REQUEST_ID(command *c, int connect_id);
void build_NEW_OBJECT(command *c, long mess_id, long ref, object_instance *object);
void build_OBJECT_CREATED(command *c, long mess_id, long ref, long s_ref);
void build_CREATION_FAILED(command *c, long mess_id, long ref, int errcode);
void build_DELETE_OBJECT(command *c, long mess_id, long ref, long r_ref);
void build_OBJECT_DELETED(command *c, long mess_id, long ref);
void build_OBJECT_LOAD(command *c, long mess_id, char *obj_name, long ref, int proc_id, long port_num);
void build_OBJECT_MESSAGE(command *c, long ref, long r_ref, long mess_id, int the_message, parameter_block *params, int wait_return);
void build_RETURN_MESSAGE(command *c, long ref, long mess_id, long result, parameter_block *params);
void build_LINK_OBJECT(command *c, long ref, long s_ref);
void build_SPEED_LINK(command *c, long ref);
void build_SPEED_MADE(command *c, int status, long ref, long port_num);
void build_OBJECT_LINKED(command *c, int status, long ref, object_instance *object);

void build_GET_CLASS_INFO(command *c, long messid, long ref, int control, int num, char *class);
void build_R_CLASS_INFO(command *c, long messid, long ref, int num, char **machines);
void build_REGISTER_SERV(command *c, char *type, char *machine, long port_num);
void build_NO_OP(command *c);
void build_LAUNCH_SERVICE(command *c, int mess_id, char *service, char *machine);
void build_START_TIME(command *c, long recv_ref);
void build_GET_TIME(command *c, long messid, long recv_ref, long send_ref);
void build_SEND_TIMES(command *c, long messid, long ref, double time_idle, double time_cpu, double time_writing);
void build_QUERY_SERV(command *c, long messid, char *machine, char *service);
void build_SERV_INFO(command *c, long messid, char *machine, char *service, long port_num, int running);

#else
void write_params();
void write_object();
void write_node_des();
void write_link_des();
int read_object();
int read_node_des();
int read_link_des();
int read_params();
command *new_command();
int read_command();
int write_command();
void clean_command();
void free_command();
int complete_command();
int complete_a_command();
void open_commands();
void close_commands();
void build_REQUEST_ID();
void build_NEW_OBJECT();
void build_OBJECT_CREATED();
void build_CREATION_FAILED();
void build_DELETE_OBJECT();
void build_OBJECT_DELETED();
void build_OBJECT_LOAD();
void build_OBJECT_MESSAGE();
void build_RETURN_MESSAGE();
void build_LINK_OBJECT();
void build_SPEED_LINK();
void build_SPEED_MADE();
void build_OBJECT_LINKED();
void build_GET_CLASS_INFO();
void build_R_CLASS_INFO();
void build_REGISTER_SERV();
void build_NO_OP();
void build_LAUNCH_SERVICE();
void build_START_TIME();
void build_GET_TIME();
void build_SEND_TIMES();
void build_QUERY_SERV();
void build_SERV_INFO();
#endif



