/* List.c
 * Jim Gochee              
 * 2/6/89
 * MODIFIED 3/9/92 with new interface style
 *
 * This file contains routines which manage a list structure called 'List'.
 * A 'List' is a linked list with forward and backward pointers. Pointers
 * to the current place in the list are kept internally, and all function
 * calls which deal with a list position use this internal pointer. The pointer
 * can be set with the function calls: SetBeginning, SetEnd, SetNext, SetPrev,
 * SetKey.
 */

#include <string.h>
#include "List.h"

#define TRUE -1
#define FALSE 0


/*****************************************************************************
 *                                                                           *
 *                            Global Functions                               *
 *                                                                           *
 *****************************************************************************/


/* CREATELIST
 *
 * This function returns a newly created list 
 */
List CreateList()
{
  List mylist;

  mylist = (List) malloc(sizeof (ListRec));
  mylist->dummy = (Nodeptr) malloc(sizeof (NodeRec));

  mylist->dummy->next = mylist->dummy;
  mylist->dummy->prev = mylist->dummy;
  mylist->current = mylist->dummy;
  mylist->num_nodes = 0;
  return(mylist);
}


/* SETBEGINNING
 *
 * This function sets internal node pointer to the beginning of the list
 */
void SetBeginning(list)
     List list;
{
  list->current = list->dummy;
}


#ifndef EndList
/* ENDLIST
 *
 * This function sets the internal node pointer to the end of the list
 */
int EndList(list)
     List list;
{
  return(list->current->next == list->dummy);
}
#endif


/* BEGINNINGLIST
 *
 * This function returns TRUE if the internal node pointer is at the beginning
 * of the list, and FALSE if it is at the end of the list
 */
int BeginningList(list)
     List list;
{
  return(list->current == list->dummy);
}


/* SETNEXT
 *
 * This function sets the internal node pointer to the next node. If we are at
 * the end of the list, then we stay there.
 */
void SetNext(list)
     List list;
{
  if (!EndList(list))
    list->current = list->current->next;
}


/* SETPREV
 *
 * This function sets the internal node pointer to the previous node. If we
 * are at the beginning of the list, then we stay there.
 */
void SetPrev(list)
     List list;
{
  if (!BeginningList(list))
    list->current = list->current->prev;
}


/* SETEND
 *
 * This routine sets the internal node pointer to the end of the list
 */
void SetEnd(list)
     List list;
{
  list->current = list->dummy->prev;
}


/* SETKEY
 *
 * This function attempts to set to the node with key equal to 'key'. If the
 * routine succeeds, TRUE is returned, otherwise FALSE is returned.
 */
int SetKey(list, key)
     List list;
     void *key;
{
  void  *tempkey;

  SetBeginning(list);
  while(!EndList(list)) {
    ReadData(list, &tempkey, 0);
    if (tempkey == key)
      return(TRUE);
    SetNext(list);
  }
  return(FALSE);
}

/* SETNUMBER
 *
 * This routines sets the list to the 'n'th node. This is an inefficient
 * function since it involves stepping through n nodes to get to the correct
 * spot. The routine returns the actual position stopped at, in case the list
 * is smaller than the given 'n'.
 */
SetNumber(list, n)
     List list;
     int n;
{
  int count=1;

  n--;
  SetBeginning(list);
  while(!(EndList(list)) && n>0) {
    count++; n--;
    SetNext(list);
  }
  return(count);
}


/* READDATA
 *
 * This routine reads the key and data from the current node. If either key
 * or data parameters to the function are 0, then the respective variable is
 * not set.
 */
ReadData(list, key,data)
     List list;
     void **key, **data;
{
  if(EndList(list)) return(FALSE);

  if (key) 
    *key = list->current->next->key;
  if (data) 
    *data = list->current->next->data;
  return(TRUE);
}


/* WRITEDATA
 *
 * This routine writes a key and data pair to the given node. If the current
 * node is the end of the list, then no writing takes place and FALSE is
 * returned.
 */
WriteData(list, key, data)
     List list;
     void *key,*data;
{
  if(EndList(list)) return(FALSE);
  list->current->next->key = key;
  list->current->next->data = data;
  return(TRUE);
}


/* INSERTDATA
 *
 * This function creates a new node and stores 'key' and 'data'.
 */
Nodeptr InsertData(list, key, data)
     List list;
     void *key, *data;
{
  Nodeptr nextnode = list->current->next;
  list->current->next = (Nodeptr) malloc(sizeof (NodeRec));
  
  list->current->next->prev = list->current;
  list->current->next->next = nextnode;
  nextnode->prev = list->current->next;
  list->current->next->key = key;
  list->current->next->data = data;
  list->num_nodes++;
  
  return(list->current->next);
}

/* SETNODE
 *
 */
void SetNode(list, n)
     List list;
     Nodeptr n;
{
  list->current = n->prev;
}


/* GETNODE
 *
 */
Nodeptr GetNode(list)
     List list;
{
  return(list->current);
}


/* DELETENODE
 *
 * This routine deallocates the memory used for the node and realigns the list
 * to reflect the change.
 */
void DeleteNode(list)
     List list;
{
  Nodeptr tempnode = list->current->next;

  if (!EndList(list)) {
    list->current->next = list->current->next->next;
    list->current->next->prev = list->current;

    free(tempnode);
    list->num_nodes--;
  }
}


/* FREELIST
 *
 * This routine calls 'DeleteNode' if there are still nodes in the list, then
 * it releases memory associated with the list.
 */
void FreeList(list)
     List list;
{
  SetBeginning(list);
  while(!EndList(list))
    DeleteNode(list);
  free(list->dummy);
}


/* NUMBER
 *
 * Returns the number of nodes in the given list
 */
int Number(list)
     List list;
{
  return(list->num_nodes);
}





