/* OBJLIB.C
 * Jim Gochee
 * 5/14/92
 *
 * This file is the glue between objects in SPEDE and the rest of the SPEDE
 * system.
 */
#include <netdb.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <varargs.h>
#include "List.h"

#include "reglib.h"
#include "select.h"
#include "sockets.h"
#include "network.h"
#include "spede.h"
#include "objects.h"
#include "spede_io.h"
#include "objlib.h"
#include "misc.h"


#define TRUE  1
#define FALSE 0

/*#define DEBUG*/


typedef struct _response_ent {
  void (*callback)();
  void *callback_data;
  long message_id;
} response_ent;


/* Extra info that goes with each object */

typedef struct _object_ent {
  object_instance *object;
  unsigned char mess_counter;
  int socket;                              /* The socket to that dean */
  List message_queue;
  void *data;                              /* Current State */
  long len;                                /* length of state data */
  int is_fault_tolerant;
  List fault_m_queue;                      /* Messages since last save */
  int num_since_save;                      /* Num messages since last save */
  int total_messages;                      /* Total number of messages */
  int sent_return_state;
  Nodeptr nodeptr;
  long dean_ref;                           /* The ref num for the dean */
  
  List pending_responses;                   /* Messages waiting to return */
  int created_async;
} object_ent;


/* Used to keep track of which deans we are connected to */

typedef struct _dean_ent {
  node_des dean_machine;
  unsigned char connect_id;
  int socket;
  connect_entry *ce;
  int num_links;
  Nodeptr nodeptr;
} dean_ent;


typedef struct {
  long message_id;
  object_ent *ent;
  return_info *r_info;
} r_des_type;


extern int errno;
char *my_optarg;
extern int optind, opterr;

/***************************************************************************
 *                                                                         *
                             Global Variables
 *                                                                         *
 ***************************************************************************/

List objects;                           /* List of current objects */
List dean_list;                         /* List of dean connections */
List create_list;
List return_list;
connect_type *dean_connections;         /* List of dean sockets to watch */
node_des local_node;                    /* The machine we are local to */
int local_id = 1;                       /* Used to make part of obj id */
void (*debug_func)();                   /* Callback for debugging commands */
long (*obj_callback)();                 /* The main object callback */
object_instance *main_object;           /* The main object descriptor */
long main_ref;                          /* Reference of the main object */
int world_port, world_connect;          /* Used to open our own listener */
extern char *object_name;
int service_socket;
double time_computing=0.0, time_idle=0.0, time_writing=0.0;
int _argc;
char **_argv;
object_descriptor run_info;
int SpedeError;
char *base_dir;
int reg_port;
char reghost[HOST_NAME_SIZE];


/***************************************************************************
 *                                                                         *
                              Local prototypes
 *                                                                         *
 ***************************************************************************/

#ifdef __STDC__
long invoke_object(long (*func)(), int mess, pblock *params, 
		    pblock **r_params);
void obj_message(command *comm);
int receive_a_command(int socket);
void socket_went_down(connect_entry *ce);
object_ent *add_object_list(object_instance *object, int socket);
void delete_object_list(object_instance *object);
long new_message_id(object_instance *);
void delete_main_object(command *);
#endif

/***************************************************************************
 *                                                                         *
                             Object Functions
 *                                                                         *
 ***************************************************************************/


void log(format, va_alist)
     char *format;
     va_dcl
{
  va_list ap;
  
#ifdef DEBUG
  va_start(ap);
  vfprintf(stdout, format, ap);
  va_end(ap);
#endif
}


/* ADD_OBJECT_LIST
 * 
 * This routine adds an object to the list of current objects. The current
 * objects are: the main object plus objects that the main object creates.
 * The return value is a reference number pointing to the object entry
 * structure. The reference number is given to the dean and all communication
 * will use the number.
 */
object_ent *add_object_list(object, socket)
     object_instance *object;
     int socket;
{
  object_ent *ent;

  ent = (object_ent *) malloc(sizeof(object_ent));
  ent->object = object;
  ent->message_queue = CreateList();        /* List of queued in messages */
  ent->fault_m_queue = CreateList();        /* List of remote out messages */
  ent->pending_responses = CreateList();
  ent->num_since_save = 0;                  /* Num messages since save */
  ent->total_messages = 0;                  /* Total messages since creation */
  ent->is_fault_tolerant = object->is_fault_tolerant;
  ent->len = 0;
  ent->socket = socket;                     /* Socket for out communication */
  ent->mess_counter = 0;                    /* Used to number out messages */
  object->data = (void *) ent;              /* Connects object to entry */

  ent->nodeptr = InsertData(objects, 0, (void *) ent);
  return(ent);                       /* Return the reference number */
}


/* DELETE_OBJECT_LIST
 *
 * This routine is called when an object we created is deleted. Note that the
 * main object can never be deleted from the list, because when the main
 * object is deleted, the process completes
 */
void delete_object_list(object)
     object_instance *object;
{
  object_ent *ent;

  ent = object->data;
  SetNode(objects, ent->nodeptr);
  DeleteNode(objects);
  free(ent);
}


/* NEW_MESSAGE_ID
 *
 * This routine generates a new message id based on the given object.
 */
long new_message_id(object)
     object_instance *object;
{
  long temp_long;
  object_ent *ent = (object_ent *) object->data;

  temp_long = object->id + (((long) ent->mess_counter) << 24);
  ent->mess_counter = (ent->mess_counter + 1) % 256;

  return(temp_long);
}



/***************************************************************************
 *                                                                         *
                              MISC FUNCTIONS
 *                                                                         *
 ***************************************************************************/


/* INVOKE_OBJECT
 *
 * This function invokes the object specified by a callback function (func)
 * with the given message (mess) and parameters (params). The object can
 * return a long value along with optional return parameters (r_params).
 */
long invoke_object(func, mess, params, r_params)
     long (*func)();
     int mess;
     pblock *params, **r_params;
{
  pblock temp_p;
  long result;

  if (params == 0) {
    temp_p.num_params = 0;
    params = &temp_p;
  }

  result = (*func)(mess, params, r_params);
  return(result);
}

  
/* OBJ_MESSAGE
 *
 * This function calls the main object with the specified message and params.
 */
void obj_message(comm)
     command *comm;
{
  pblock *r_params;
  command s_mess;
  object_ent *ent;
  long result;
  double start_time, start_sending;

  ent = (object_ent *) comm->recv_ref;

  start_time = get_seconds();
  result = invoke_object(obj_callback, comm->data.mess.the_message, 
			 comm->data.mess.params, &r_params);
  time_computing += (get_seconds() - start_time);

  start_time = get_seconds();
			   
  /* Send back the results */

  if (comm->data.mess.wait_return) {
    build_RETURN_MESSAGE(&s_mess, comm->send_ref, comm->message_id, result, r_params);
    log("(%s) sending return message... to ref %d socket %d\n", object_name, comm->send_ref, comm->socket);
    start_sending = get_seconds();
    write_command(&s_mess, comm->socket);
    log("Done sending return values. Time taken = %.10g seconds.\n", get_seconds() - start_sending);
  }
  time_writing += (get_seconds() - start_time);
  clean_params(r_params);
}


/* DELETE_MAIN_OBJECT
 *
 * This function is called when the delete command is received. The object
 * kills itself then the process exits.
 */
void delete_main_object(comm)
     command *comm;
{
  command s_comm;
  long dean_ref;

  dean_ref = ((object_ent *) main_object->data)->dean_ref;

  invoke_object(obj_callback, DELETE, 0, 0);    /* Allow object to clean up */

  build_OBJECT_DELETED(&s_comm, comm->message_id, comm->send_ref);
  write_command(&s_comm, comm->socket);
  log("(%s) - Terminating naturally\n", object_name);
  exit(0);
}


void form_a_link(comm)
     command *comm;
{
  command s_comm;
  object_ent *ent;

  log("OBJLIB - another object wants to link with us\n");

  ent = (object_ent *) comm->recv_ref;
  build_OBJECT_LINKED(&s_comm, 0, (long) ent, ent->object);
  write_command(&s_comm, comm->socket);
}  
  

int handle_callback(comm)
     command *comm;
{
  object_ent *ent;
  response_ent *re;
  return_info *r_info;
  r_des_type *return_des;

  ent = (object_ent *) comm->recv_ref;
  
  SetBeginning(ent->pending_responses);
  while(!EndList(ent->pending_responses)) {
    ReadData(ent->pending_responses, 0, (void **) &re);
    if (re->message_id == comm->message_id) {
      DeleteNode(ent->pending_responses);
      r_info = new_rinfo();
      r_info->status = OBJECT_OK;
      r_info->result = comm->data.return_m.result;
      r_info->r_params = comm->data.return_m.params;
      (*re->callback)((object_instance *) ent->object, (return_info *) r_info, 
		      (void *) re->callback_data);
      free(re);
      free(comm);                         /* Note that we free the command */
      return(TRUE);
    }
    SetNext(ent->pending_responses);
  }
  
  /* See if the return_message is for an asynchronous call */

  SetBeginning(return_list);
  while(!EndList(return_list)) {
    ReadData(return_list, 0, (void **) &return_des);
    if (return_des->ent == ent && 
	return_des->message_id == comm->message_id){
      r_info = return_des->r_info;
      r_info->status = OBJECT_OK;
      r_info->result = comm->data.return_m.result;
      r_info->r_params = comm->data.return_m.params;
      free(comm);
      return(TRUE);
    }
    SetNext(return_list);
  }
  return(FALSE);
}


void get_times(comm)
     command *comm;
{
  command s_comm;

  build_SEND_TIMES(&s_comm, comm->message_id, comm->send_ref, time_idle, time_computing, time_writing);

  write_command(&s_comm, comm->socket);
}


void object_created_async(comm)
     command *comm;
{
  object_ent *t_ent, *ent;

  ent = (object_ent *) comm->recv_ref;
  SetBeginning(create_list);
  while(!EndList(create_list)) {
    ReadData(create_list, 0, (void **) &t_ent);
    if (t_ent == ent) {
      t_ent->object->state = OBJECT_OK;
      ent->dean_ref = comm->send_ref;
      ent->object->location.obj_ref = ent->dean_ref;
      free_command(comm);
      return;
    }
    SetNext(create_list);
  }
  printf("Warning!!! An object was created asynchronously without my knowledge\n");
}


void queue_message(new)
     command *new;
{
  object_ent *obj_ent;

  obj_ent = (object_ent *) new->recv_ref;
  SetEnd(obj_ent->message_queue);
  InsertData(obj_ent->message_queue, 0, (void *) new);
}

/* OBJECT_COMMAND
 *
 * This function is called when a socket has some data. It is assumed that the
 * data will be in the command format. Commands are then processed accordingly
 */
int object_command(new)
     command *new;
{
  object_ent *obj_ent;
  
  log("(%s) - queuing command '%s'\n", object_name, command_name[new->comm]);
  switch(new->comm) {
  case START_TIME:
    time_computing = 0.0;
    time_idle = 0.0;
    time_writing = 0.0;
    break;
  case GET_TIME:
    get_times(new);
    break;
  case LOG_MESSAGE:
    if (debug_func)
      (*debug_func)(new->data.log.comm); 
    free_command(new);
    return(0);
  case OBJECT_CRASHED:
    return(0);
  case STATE_INFO:
    return(0);
  case NO_OP:
    return(0);
  case RETURN_MESSAGE:
    if (handle_callback(new))
      return(0);
    /* Else fall through to queuing the message below */
    queue_message(new);
    break;
  case SEND_TIMES:
    queue_message(new);
    break;
  case OBJECT_MESSAGE:                          /* Queue this command */
    queue_message(new);
    break;
  case OBJECT_CREATED:                          /* Queue this command */
    obj_ent = (object_ent *) new->recv_ref;
    if (obj_ent->created_async) {
      object_created_async(new);
      return(0);
    }
    queue_message(new);
    break;
  case OBJECT_DELETED:                          /* Queue this command */
    queue_message(new);
    break;
  case R_CLASS_INFO:
    return(0);
  case DELETE_OBJECT:
    delete_main_object(new);
    break;
  case LINK_OBJECT:
    form_a_link(new);
    break;
  default:
    printf("Unknown command %s\n", command_name[new->comm]);
    break;
  }
}


/* SOCKET_WENT_DOWN
 * 
 * This function is called when a socket has gone bad.
 */
void socket_went_down(ce)
     connect_entry *ce;
{
  close_commands(ce->the_socket);
  delete_connection(dean_connections, ce);
  log("Socket %d went down\n", ce->the_socket);
  log("(%s) - Exiting\n", object_name);
  exit(1);
}


/* IDLE_OBJECTS
 *
 * This routine should be called as often as possible. It checks the network
 * and handles any I/O processing
 */
void idle_objects(secs, usecs)
     int secs, usecs;
{
  int nfound;
  struct timeval ti;
  connect_entry *ce;
  double start_time;
  int status;
  command *comm;

  ti.tv_sec = secs;
  ti.tv_usec = usecs;

  start_time = get_seconds();
  nfound = wait_on_connections(dean_connections, &ti);
  time_idle += (get_seconds() - start_time);

  if (nfound) {
    while(ce = get_next_entry(dean_connections)) {
      if (ce->the_socket == world_connect) {
	log("OBJLIB - Accepting a new connection\n");
	ce = accept_new_connection(dean_connections, world_connect);
	open_commands(ce->the_socket);
      }
      else {
	/* It is a command */
	/*printf("Reading from socket %d\n", ce->the_socket); */
	log("OBJLIB - Data on socket %d\n", ce->the_socket);
	if ((status = complete_a_command(&comm, ce->the_socket))) {
	  if (status == -1)
	    socket_went_down(ce);
	  else {
	    object_command(comm);
	  }
	}
	else {
	}
	/* printf("Done reading from socket %d\n", ce->the_socket); */
      }
    }
  }
}



int sync_created_objects()
{
  object_ent *ent;
  int done;

  while(TRUE) {
    done = TRUE;
    SetBeginning(create_list);
    while(!EndList(create_list)) {
      ReadData(create_list, 0, (void ** )&ent);
      if (ent->object->state == CREATING_OBJECT) {
	done = FALSE;
	break;
      }
      SetNext(create_list);
    }
    if (done)
      return(0);
    idle_objects(2,0);
  }
}     
      
int get_object_status(object)
     object_instance *object;
{
  return(object->state);
}


/* OBJECT_HAS_MESSAGE
 *
 * This routine checks to see if a given object (object) has a message that
 * messages a given id (message_id) or a command (comm_type).
 * If there exists such a message, then a command pointer is returned. If not,
 * 0 is returned.
 */
command *object_has_message(object, message_id)
     object_instance *object;
     long message_id;
{
  command *comm;
  object_ent *obj_ent;
  List l;

  idle_objects(2,0);                            /* Allow for network I/O */
  obj_ent = (object_ent *) object->data;
  l = obj_ent->message_queue;
  
  SetBeginning(l);
  while(!EndList(l)) {
    ReadData(l, 0, (void **) &comm);
    if (comm->message_id == message_id) {
      DeleteNode(l);
      log("(%s) - De-queued command  %s\n", object_name, command_name[comm->comm]); 
      return(comm);
    }
    SetNext(l);
  }
  return(0);
}



command *object_has_a_mess(object, message_type)
     object_instance *object;
     long message_type;
{
  command *comm;
  object_ent *obj_ent;
  List l;

  idle_objects(0, 3);                            /* Allow for network I/O */
  obj_ent = (object_ent *) object->data;
  l = obj_ent->message_queue;
  
  SetBeginning(l);
  while(!EndList(l)) {
    ReadData(l, 0, (void **) &comm);
    if (comm->comm == message_type) {
      DeleteNode(l);
      log("(%s) - De-queued command  %s\n", object_name, command_name[comm->comm]); 
      return(comm);
    }
    SetNext(l);
  }
  return(0);
}



dean_ent *get_remote_dean(host_node, dean_socket, connect_id)
     node_des *host_node;
     int *dean_socket;
     unsigned char *connect_id;
{
  dean_ent *ent;
  int socket, result;
  short running;
  command comm;
  long port_num;

  SetBeginning(dean_list);
  while(!EndList(dean_list)) {
    ReadData(dean_list, 0, (void **) &ent);
    if (!strcmp(ent->dean_machine.name, host_node->name)) {
      log("OBJLIB - Get remote dean found a connection...\n");
      *dean_socket = ent->socket;
      ent->num_links++;
      if (connect_id)
	*connect_id = ent->connect_id;
      return(ent);
    }
    SetNext(dean_list);
  }

  /* Need to create a connection */  
  
  query_server(host_node->name, "dean", &running, &port_num);
  if (!running) {
    log("Telling Reg to launch dean on %s\n", host_node->name);
    if ((result = launch_service(0, host_node->name, "dean", &port_num))) {
      if (result == -1)
	log("That launch timed-out\n");
      else
	log("There was a socket read error\n");
      *dean_socket = -1;
      return((dean_ent *) 0);
    }
  }
  
  socket = connect_remote(port_num, host_node->addr, 1);
  if (socket == -1) {
    log("Was told dean was running, but apparently this is not true\n");
    *dean_socket = -1;
    return((dean_ent *) 0);
  }
  
  build_REQUEST_ID(&comm, 0);
  write_command(&comm, socket);
  read_command(&comm, socket);

  if (connect_id)
    *connect_id = comm.data.req_id.connect_id;
  *dean_socket = socket;
  
  ent = (dean_ent *) malloc(sizeof(dean_ent));
  copy_node_des(&ent->dean_machine, host_node);
  ent->connect_id = comm.data.req_id.connect_id;
  ent->socket = socket;
  ent->num_links = 1;
  ent->nodeptr = InsertData(dean_list, 0, (void *) ent);

  ent->ce = add_connection(dean_connections, socket, host_node);
  open_commands(socket);

  return(ent);
}


void delete_dean_ent(ent)
     dean_ent *ent;
{
  if (ent->num_links > 1) {
    ent->num_links--;
    return;
  }

  SetNode(dean_list, ent->nodeptr);
  DeleteNode(dean_list);
  delete_connection(dean_connections, ent->ce);
  free(ent);
}

  


/* CREATE_OBJECT
 *
 * This routine is called by the user to create a new remote object. The
 * first parameter asks for a machine name in standard unix format, such as
 * "coos.dartmouth.edu". If instead of a machine, NULL is passed, the local
 * machine will be used. The second parameter is the type of object, described
 * as an ascii string. The third paramter is a time quanta used to determine
 * how often the object should be checked for moving. The last two parameters
 * control whether or not an object is debugged and 'logged' for fault
 * tolerance. Both options significantly decrease object performance.
 */
object_instance *create_object(machine, type, debug, fault_tol, create_async)
     char *machine, *type;
     int debug, fault_tol, create_async;
{
  object_instance *object;
  struct hostent *hp;
  node_des remote_node;
  int dean_socket;
  unsigned char connect_id;
  int mess_id;
  command comm, *ret_mes;
  object_ent *ent;

  if (machine == 0)
    copy_node_des(&remote_node, &local_node);
  else {
    hp = gethostbyname(machine);
    if (hp == 0) {
      printf("Illegal host name %s\n", machine);
      return(0);
    }
    hent_to_ndes(hp, &remote_node);
    strcpy(remote_node.name, machine);
  }
  
  object = new_object();                           /* Make the object */
  
  /* The main object is the creator of the new object */
  copy_node_des(&object->creator.node, &local_node);
  object->creator.obj_ref = main_object->dean_ref;

  /* The main object is the storesite of the main object */
  copy_node_des(&object->storesite.node, &local_node);
  object->storesite.obj_ref = main_object->dean_ref;

  /* The remote node is the location */
  copy_node_des(&object->location.node, &remote_node);

  strcpy(object->type, type);
  object->is_fault_tolerant = fault_tol;
  object->debug = debug;


  get_remote_dean(&remote_node, &dean_socket, &connect_id);
  if (dean_socket == -1) {
    free_object(object);
    printf("Could not launch remote dean.\n");
    return(0);
  }

  object->id = local_id + (((long) connect_id) << 16);
  local_id = (local_id + 1) % 16000;
  
  ent = add_object_list(object, dean_socket);
  
  mess_id = new_message_id(object);
  build_NEW_OBJECT(&comm, mess_id, (long) ent, object);
  write_command(&comm, dean_socket);

  if (create_async) {
    object->state = CREATING_OBJECT;
    ent->created_async = TRUE;
    SetEnd(create_list);
    InsertData(create_list, 0, (void *) ent);
    SpedeError = NO_ERROR;
    return(object);
  }
  else {
    ent->created_async = FALSE;
    while(!(ret_mes = object_has_message(object, mess_id)));
    object->state = OBJECT_OK;
    if (ret_mes->comm != OBJECT_CREATED) {
      log("Object not created. Command %s received\n", 
	  command_name[ret_mes->comm]);
      free_command(ret_mes);
      free_object(object);
      SpedeError = INVALID_OBJECT;
      return(0);
    }

    ent->dean_ref = ret_mes->send_ref;
    object->location.obj_ref = ent->dean_ref;
    
    free_command(ret_mes);
    SpedeError = NO_ERROR;
    return(object);
  }
}


object_instance *link_object(the_link)
     link_des *the_link;
{
  int dean_socket, connect_id;
  object_instance *object;
  command comm;
  object_ent *ent;

  get_remote_dean(&the_link->node, &dean_socket, 0);
  object = new_object();

  ent = add_object_list(object, dean_socket);
/*  log("OBJLIB link id is %d\n", the_link->obj_ref); */
  build_LINK_OBJECT(&comm, the_link->obj_ref, (long) ent);
  write_command(&comm, dean_socket);
  read_command(&comm, dean_socket);
  
  if (comm.data.o_linked.status != 0) {
    log("Link attempt failed...%d\n", comm.data.o_linked.status);
    return(0);
  }
  log("OBJLIB - Link was successful\n");
  *object = comm.data.o_linked.object;
  object->data = ent;
  ent->dean_ref = comm.send_ref;
  
  return(object);
}



object_instance *speed_link(the_link)
     link_des *the_link;
{
  int dean_socket, connect_id, socket;
  object_instance *object;
  command comm;
  object_ent *ent;
  dean_ent *dean;
  long obj_ref;

  dean = get_remote_dean(&the_link->node, &dean_socket, 0);
  object = new_object();

  ent = add_object_list(object, dean_socket);
  build_SPEED_LINK(&comm, the_link->obj_ref);
  write_command(&comm, dean_socket);
  read_command(&comm, dean_socket);
  obj_ref = comm.send_ref;

  delete_dean_ent(dean);                        /* We are done with the dean */

  if (comm.data.speed_made.status != 0) {
    log("OBJLIB - Speed link attempt failed...%d\n", comm.data.speed_made.status);
    return(0);
  }

  log("OBJLIB - Speed link was successful...make link with object\n");

  /* Make link with object directly */

  socket = connect_remote(comm.data.speed_made.port_num, the_link->node.addr, 5);

  log("OBJLIB - Connected with remote object through socket %d, port %d\n", socket, comm.data.speed_made.port_num);
  if (socket == -1) {
    log("OBJLIB - A link could not be made!!!\n");
    return(0);
  }

  build_LINK_OBJECT(&comm, obj_ref, (long) ent);
  write_command(&comm, socket);
  log("OBJLIB - waiting for object to respond\n");
  read_command(&comm, socket);
  
  if (comm.data.o_linked.status != 0) {
    log("OBJLIB - Link attempt failed...%d\n", comm.data.o_linked.status);
    return(0);
  }
  log("OBJLIB - Link was successful...link ref is %d\n", (long) ent);
  *object = comm.data.o_linked.object;
  object->data = ent;
  ent->dean_ref = obj_ref;
  ent->socket = socket;
  add_connection(dean_connections, socket, &the_link->node);
  open_commands(socket);

  return(object);
}


int send_obj_mess_callback(object, message, params, call_b, data)
     object_instance *object;
     int message;
     pblock *params;
     void *call_b, *data;
{
  response_ent *re;
  object_ent *ent;
  int mess_id;
  command comm;

  ent = (object_ent *) object->data;
  mess_id = new_message_id(object);

  build_OBJECT_MESSAGE(&comm, ent->dean_ref, (long) ent, mess_id, message, params, TRUE);

  write_command(&comm, ent->socket);

  re = (response_ent *) malloc(sizeof(response_ent));
  re->callback = call_b;
  re->message_id = mess_id;
  re->callback_data = data;
  SetEnd(ent->pending_responses);
  InsertData(ent->pending_responses, 0, (void *) re);

  return(0);
}
  


return_info *send_obj_message(object, message, params, async, wait_return)
     object_instance *object;
     int message;
     pblock *params;
     int async;
     int wait_return;
{
  long result;
  long mess_id;
  command comm, *r_comm;
  object_ent *ent;
  return_info *new_info;
  r_des_type *return_des;

  new_info = new_rinfo();
  
  ent = (object_ent *) object->data;
/*  log("Sending message %d to object %d\n", message, object->id); */
  mess_id = new_message_id(object);

  build_OBJECT_MESSAGE(&comm, ent->dean_ref, (long) ent, mess_id, message, params, wait_return);
/*  log("(%s) Sending message through dean ref %d\n", object_name, ent->dean_ref); */
  write_command(&comm, ent->socket);
  if (!async) {
    while(!(r_comm = object_has_message(object, mess_id)));
/*    log("Got back from message\n");*/
    new_info->status = OBJECT_OK;
    new_info->result = r_comm->data.return_m.result;
    new_info->r_params = r_comm->data.return_m.params;
    free(r_comm);
    return(new_info);
  }
  else {
    new_info->status = 1;
    return_des = (r_des_type *) malloc(sizeof(r_des_type));
    return_des->message_id = mess_id;
    return_des->ent = ent;
    return_des->r_info = new_info;
    InsertData(return_list, 0, (void **) return_des);
    return(new_info);
  }
}


int delete_object(object)
     object_instance *object;
{
  object_ent *ent;
  long mess_id;
  command comm, *r_comm;

  ent = (object_ent *) object->data;
  mess_id = new_message_id(object);
  build_DELETE_OBJECT(&comm, mess_id, ent->dean_ref, (long) ent);
  write_command(&comm, ent->socket);
  
  while(!(r_comm = object_has_message(object, mess_id)));
  delete_object_list(object);
  free_object(object);
  return(0);
}



void get_object_times(object, idle, cpu, writing)
     object_instance *object;
     double *idle, *cpu, *writing;
{
  object_ent *ent;
  long mess_id;
  command comm, *r_comm;

  ent = (object_ent *) object->data;
  mess_id = new_message_id(object);
  build_GET_TIME(&comm, mess_id, ent->dean_ref, (long) ent);
  write_command(&comm, ent->socket);
  
  while(!(r_comm = object_has_message(object, mess_id)));
  *idle = r_comm->data.send_times.time_idle;
  *cpu = r_comm->data.send_times.time_computing;
  *writing = r_comm->data.send_times.time_writing;

  free_command(r_comm);
}



void reset_times(object)
     object_instance *object;
{
  object_ent *ent;
  long mess_id;
  command comm, *r_comm;

  ent = (object_ent *) object->data;
  mess_id = new_message_id(object);
  build_START_TIME(&comm, ent->dean_ref);
  write_command(&comm, ent->socket);
}


int valid_option(check_char, test, expect_arg)
     char check_char;
     char *test;
     int *expect_arg;
{
  while(*test != 0 && *test != check_char)
    test++;
  if (*test == 0)
    return(FALSE);
  test++;
  if (*test == ':')
    *expect_arg = TRUE;
  else
    *expect_arg = FALSE;
  return(TRUE);
}


int my_getopt(argc, argv, test)
     int *argc;
     char **argv, *test;
{
  static int position = 1;
  int expect_arg, k, start;
  char next_opt;

  while(TRUE) {
    if (position >= *argc)
      return(EOF);
    if (argv[position][0] != '-') {
      position++;
      continue;
    }
    next_opt = argv[position][1];
    if (valid_option(next_opt, test, &expect_arg)) {
      start = position;
      position++;
      if (expect_arg) {
	my_optarg = argv[position];
	position++;
      }
      for(k=position; k < *argc; k++)
	argv[start++] = argv[k];
      *argc -= 1 + ((expect_arg) ? 1 : 0);
      position -= 1 + ((expect_arg) ? 1 : 0);
      return(next_opt);
    }
    position++;
  }
}
  

void _set_options(argc, argv)
     int *argc;
     char **argv;
{
  int ch;
  int use_file = FALSE;
  int num_machines=0, k, class_machines;
  char machines[256];
  char f_name[256];
  char **temp_machines;
  FILE *fp;

  machines[0] = '\0';
  
  /* Set preliminary info */
  run_info.num_objects = 0;
  run_info.classes[0] = '\0';
  run_info.num_machines = 0;
  run_info.machine = 0L;

  while((ch = my_getopt(argc, argv, "h:p:n:c:f:m:")) != EOF) {
    switch(ch) {
    case 'n':
      run_info.num_objects = atoi(my_optarg);
      break;
    case 'c':
      strcpy(run_info.classes, my_optarg);
      break;
    case 'm':
      strcpy(machines, my_optarg);
      break;
    case 'f':
      use_file = TRUE;
      strcpy(f_name, my_optarg);
      break;
    case 'h':
      strcpy(reghost, my_optarg);
      break;
    case 'p':
      reg_port = atoi(my_optarg);
      break;
    }
  }

  if (use_file) {
    fp = fopen(f_name, "r");
    if (fp == 0) {
      printf("ERROR!! Could not open file '%s'\n", f_name);
      use_file = FALSE;
    }
  }
  
  if (use_file) {
    fscanf(fp, "%d\n", &run_info.num_objects);
    fgets(run_info.classes, 100, fp);
    run_info.classes[strlen(run_info.classes)-1] = 0;
    fgets(machines, 100, fp);
    machines[strlen(machines)-1] = 0;
    fclose(fp);
  }
  
  
/* This was taken out 5/31/92 because the 'reg' process did not have enough
 * file descriptors available.
 */
  register_info(reghost, reg_port);

  service_socket = connect_register_serv();
  if (service_socket == -1) {
    read_register_info();
    service_socket = connect_register_serv();
    if (service_socket == -1) {
      fprintf(stderr, "Cannot connect with reg process!!!\n", reghost, reg_port);
      exit(0);
    }
  }
  

  if (run_info.classes[0] != '\0') {
    class_machines = get_machines(run_info.classes, &temp_machines);
  }
  else
    class_machines = 0;

  if (machines[0] == '\0')
    num_machines = 0;
  else
    num_machines = num_o_items('#', machines);

  if (num_machines+class_machines) {
    run_info.machine = (char **) malloc(sizeof(char *) * (num_machines + class_machines));
    for(k=1; k <= num_machines; k++)
      run_info.machine[k-1] = alloc_o_item('#', machines, k);
    for(k=num_machines; k < (num_machines+class_machines); k++)
      run_info.machine[k] = temp_machines[k-num_machines];
  }
  run_info.num_machines = num_machines + class_machines;
  if (run_info.num_objects == 0)
    run_info.num_objects = run_info.num_machines;
  free(temp_machines);
}


/* INIT_OBJECTS
 *
 * Set up networking data structures and object structures.
 */
int init_objects(argc, argv, obj_call)
     int argc;
     char **argv;
     void *obj_call;
{
  long my_addr;
  char local_hostname[33];
  

  reghost[0] = '\0';
  reg_port = 0;

  _set_options(&argc, argv);

  _argc = argc;
  _argv = argv;

  debug_func = 0L; /*dfunc; */
  obj_callback = obj_call;
  get_local_address(&my_addr);
  
  my_getnodebyaddr(my_addr, &local_node);

  log("(%s) - Local machine: '%s'(%X)\n", object_name, local_node.name, 
      ntohl(local_node.addr));

  dean_list = CreateList();                    /* List of current deans */
  objects = CreateList();                      /* List of current objects */
  create_list = CreateList();
  return_list = CreateList();
  dean_connections = new_connections();        /* Links to deans */

  random_listener(&world_connect, &world_port, 20);
  log("OBJLIB - Set up a socket %d on port %d\n", world_connect, world_port);
  add_connection(dean_connections, world_connect, &local_node);

  gethostname(local_hostname, sizeof(local_hostname));
  
  add_connection(dean_connections, service_socket, &local_node);
  open_commands(service_socket);

  if (register_service(local_hostname, object_name, world_port, &base_dir)) {
    printf("There was an error registering the object\n");
    exit(0);
  }

  _new_main();
  return(0);
}



int _new_main()
{
  short the_id;
  object_ent *ent;
  
  main_object = new_object();

  if (obj_callback)
    invoke_object(obj_callback, CREATE, 0, 0);
  
  ent = add_object_list(main_object, 0);
  main_ref = (long) ent;
  return(0);
}


int update_main(comm)
     command *comm;
{
  object_ent *ent;
  long dean_ref;

  ent = (object_ent *) main_object->data;
  *main_object = comm->data.new_o.object;
  main_object->data = ent;

  dean_ref = comm->send_ref; 
  main_object->dean_ref = dean_ref;

  comm->comm = OBJECT_CREATED;
  comm->recv_ref = dean_ref;
  comm->send_ref = main_ref;    /* My ref num!!! */
  
  write_command(comm, comm->socket);
  ent->dean_ref = dean_ref;
  ent->socket = comm->socket;
  
  return(0);
}


void loop_forever()
{
  command *r_comm;

  /* Main object sits here and waits for outside commands */
  
  log("(%s) - Loop forevers on commands\n", object_name);
  while(TRUE) {
    while(!(r_comm = object_has_a_mess(main_object, OBJECT_MESSAGE)));
    /* log("(%s) - Handle object message\n", object_name); */
    obj_message(r_comm);
    free_command(r_comm);
  }
}


int register_dean()
{
  long my_addr;
  unsigned char connect_id;
  int dean_socket;
  command comm;
  extern char *object_name;
  extern int errno;
  char **argv;


  get_local_address(&my_addr);

  get_remote_dean(&local_node, &dean_socket, &connect_id);
  if (dean_socket == -1) {
    log("OBJLIB - Local dean not found!!\n");
    exit(0);
  }

  build_OBJECT_LOAD(&comm, 0, object_name, main_ref, (int) getpid(), world_port);
  write_command(&comm, dean_socket);  
  read_command(&comm, dean_socket);        /* Dean tells us if we are new */

  if (comm.comm == OBJECT_LOADED) {
    main_object->dean_ref = comm.send_ref;
    free_launch(_argc, _argv);
  }
  else {
    normal_launch();                       /* Give user chance to init     */
    update_main(&comm);                    /* Dean is launching new object */
    loop_forever();
  }

  return(0);
}
