/* REGLIB.C
 * Jim Gochee
 * 5/14/92
 *
 * This file is the interface to the REG process.
 */
#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "List.h"
#include "select.h"
#include "sockets.h"
#include "network.h"
#include "spede.h"
#include "spede_io.h"

#define TRUE 1
#define FALSE 0

int reg_socket = -1;
int loaded_config = FALSE;
char reg_machine[256];
long reg_port = 0;


/* REGISTER_INFO
 *
 * This function should be called to set the two crucial elements used to
 * connect with the REG process, the hostname and port number. If these are
 * not set, the connect function will fail.
 */
void register_info(host, port)
     char *host;
     int port;
{
  strcpy(reg_machine, host);
  reg_port = port;
}


/* CONNECT_REGISTER_SERV
 *
 * This routine is called when a DOPE entity wants to connect with the REG
 * process. If the connection fails, a -1 will be returned. If the connection
 * succeeded, a socket number will be returned. 
 */
int connect_register_serv()           /* returns a socket connection */
{
  struct hostent *hp;

  if ((reg_port < 1 || reg_machine[0] == '\0')) {
    return(-1);
  }
  
  hp = gethostbyname(reg_machine);
  if (hp == 0) {                      /* Bad host name */
    return(-1);  
  }

  reg_socket = connect_remote(reg_port, *(long *)hp->h_addr, 1);
  return(reg_socket);
}


/* REGISTER_SERVICE
 *
 * This routine is called right after connect_register_serv. It tells the
 * REG process that we exist, plus it allows us to read in our base directory.
 * If the routine is successful, 0 is returned. If an error occurs, -1 is
 * returned.
 */
int register_service(machine, service, port, base_dir)
     char *machine;
     char *service;
     long port;
     char **base_dir;
{
  command comm;
  int err;

  if (reg_socket == -1)
    return(-1);

  build_REGISTER_SERV(&comm, service, machine, port);
  write_command(&comm, reg_socket);
  err = read_command(&comm, reg_socket);
  if (err) {
    return(-1);
  }
  
  if (base_dir) {
    *base_dir = (char *) malloc(strlen(comm.data.base_dir.base)+1);
    strcpy(*base_dir, comm.data.base_dir.base);
  }
  return(0);
}


/* GET_MACHINES
 *
 * This routine will query the REG process for a list of machines matching
 * a class descriptor. If the routine is successful, the number of machines
 * found will be returned. If an error occurs, -1 will be returned.
 */
int get_machines(class, machine_list)
     char *class;
     char ***machine_list;
{
  command comm;
  int err;

  if (reg_socket == -1) {
    return(-1);
  }
    
  build_GET_CLASS_INFO(&comm, 0, 0, 1, 32767, class);
  write_command(&comm, reg_socket);
  err = read_command(&comm, reg_socket);
  if (err)
    return(err);

  *machine_list = comm.data.r_class.machine;
  return(comm.data.r_class.num_machines);
}


/* QUERY_MACHINES
 *
 * This routine will query the REG process for the port_num of a remote
 * entity (like a dean). If the remote entity is running, REG will return
 * its port number. If not, 'running' will be set to false.
 */
int query_server(machine, service, running, port_num)
     char *machine, *service;
     short *running;
     long *port_num;
{
  command comm;
  int err;
  
  if (reg_socket == -1)
    return(-1);

  build_QUERY_SERV(&comm, 0, machine, service);
  write_command(&comm, reg_socket);
  err = read_command(&comm, reg_socket);
  if (err) {
    *running = FALSE;
    *port_num = -1;
    return(-1);
  }
  *running = comm.data.serv_info.running;
  *port_num = comm.data.serv_info.port_num;
  return(0);
}


/* LAUNCH_SERVICE
 *
 * This routine allows the object to tell REG to launch a remote entity
 * (like a dean). The call will block until the remote entity has been
 * launched.
 */
int launch_service(id, machine, service, port_num)
     int id;
     char *machine, *service;
     int *port_num;
{
  command comm;
  int err;

  build_LAUNCH_SERVICE(&comm, id, service, machine);
  write_command(&comm, reg_socket);
  err = read_command(&comm, reg_socket);
  if (err) {
    *port_num = -1;
    return(1);
  }
  *port_num = comm.data.serv_info.port_num;

  /* Was there a problem starting the service on the remote machine? */
  
  if (!comm.data.serv_info.running) {
    *port_num = -1;
    return(-1);
  }
  else
    return(0);
}


/* READ_REGISTER_INFO
 *
 * This routine will attempt to read the REG info file. This is done when the
 * user does not specify a REG from the command line. When a REG starts up,
 * it writes this information to the user's home directory.
 */
void read_register_info()
{
  FILE *fp;
  char fname[256], *home_path;

  home_path = (char *) getenv("HOME");
  if (strcmp(home_path, ""))
    strcat(home_path, "/");             /* Only add backslash if path exists */
  
  sprintf(fname, "%s.reginfo", home_path);
  fp = fopen(fname, "r");
  if (fp == 0) {
    reg_machine[0] = '\0';
    reg_port = 0;
    return;
  }

  fgets(reg_machine, 32, fp);
  reg_machine[strlen(reg_machine)-1] = '\0';
  fscanf(fp, "%d", &reg_port);
  fclose(fp);
}
  


