/* SPEDE_IO.C
 * Jim Gochee
 * 5/14/92
 *
 * This file is responsible for medium level network protocol management. All
 * communication in SPEDE is done via the command structure. It is the task
 * of this file to read and write commands.
 */

#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>

#include "rw_sock.h"
#include "sockets.h"
#include "spede.h"
#include "spede_io.h"
#include "objects.h"

#define NEW_IO                            /* Allows chunks to be buffered */

typedef struct {
  int bytes_needed;
  int bytes_read;
  int size_of_buff;
  void *data;
} comm_ent;

#define DEFAULT_BUFF_SIZE (sizeof(command))


char *command_name[] = {"PROTO_ERROR","OK","NEW_OBJECT","CREATION_FAILED",
			  "OBJECT_CREATED","OBJECT_LOAD","OBJECT_MESSAGE",
			  "RETURN_MESSAGE","DELETE_OBJECT","OBJECT_DELETED",
			  "REQUEST_ID","LOG_MESSAGE","RETURN_STATE",
			  "STATE_INFO","OBJECT_CRASHED","SET_OBJ_STATE",
			  "FREE_LAUNCH","LINK_OBJECT","OBJECT_LOADED",
			  "OBJECT_LINKED", "SPEED_LINK", "SPEED_MADE", "KILL",
			  "REGISTER_SERV", "NO_OP", "LAUNCH_REMOTE",
			  "GET_CLASS_INFO", "R_CLASS_INFO", "LAUNCH_SERVICE",
			  "START_TIME", "GET_TIME", "SEND_TIMES","QUERY_INFO",
			  "SERV_INFO","BASE_DIR"};


long current_time = 0;
comm_ent *comm_array[MAX_SOCKETS];


/* WRITE_PARAMS
 *
 * This routine writes the parameter_block structure
 */
void write_params(p)
     parameter_block *p;
{
  int k;

  if (p == 0) {
    write_short(0);
    return;
  }
    
  write_short(p->num_params);
  for(k=0; k < p->num_params; k++) {
    write_long(p->param[k].len);
    write_block(p->param[k].data, p->param[k].len);
  }
}




/* WRITE_NODE_DES
 *
 * This routine writes a node descriptor
 */
void write_node_des(n)
     node_des *n;
{
  write_string(n->name);
  write_long(n->addr);
}


/* WRITE_LINK_DES
 *
 * This routine writes a link descriptor
 */
void write_link_des(link)
     link_des *link;
{
  write_node_des(&link->node);
  write_long(link->obj_ref);
}


/* WRITE_OBJECT
 *
 * This routine writes an object
 */
void write_object(object)
     object_instance *object;
{
  write_link_des(&object->creator);
  write_link_des(&object->storesite);
  write_link_des(&object->location);
  write_string(object->type);
  write_long(object->id);
  write_char(object->is_fault_tolerant);
  write_char(object->debug);
}


/* READ_PARAMS
 * 
 * This routine reads the parameter_block structure
 */
int read_params(socket, p)
     int socket;
     parameter_block **p;
{
  int k, err;
  short num;
  long len;

  err = read_short(socket, &num);             /* Get number of params */
  if (num == 0) {
    *p = 0;
    return(err);
  }

  *p = new_params(0);
  (*p)->num_params = num;
  for(k=0; k < num && !err; k++) {
    err = read_long(socket, &len);
    (*p)->param[k].len = len;
    (*p)->param[k].data = (void *) malloc(len);
    err |= read_block(socket, (*p)->param[k].data, len);
  }
  if (err)
    free_params(*p);
  return(err);
}


/* READ_NODE_DES
 *
 * This routine reads a node_des
 */
int read_node_des(socket, n)
     node_des *n;
{
  int err = 0;

  err |= read_string(socket, n->name, sizeof(n->name));
  err |= read_long(socket, &n->addr);
  return(err);
}


/* READ_LINK_DES
 *
 * This routine reads a link_des
 */
int read_link_des(socket, link)
     int socket;
     link_des *link;
{
  int err = 0;

  err |= read_node_des(socket, &link->node);
  err |= read_long(socket, &link->obj_ref);

  return(err);
}


/* READ_OBJECT
 *
 * This routine reads an object
 */
int read_object(socket, object)
     object_instance *object;
{
  int err=0;

  err |= read_link_des(socket, &object->creator);
  err |= read_link_des(socket, &object->storesite);
  err |= read_link_des(socket, &object->location);
  err |= read_string(socket, object->type, sizeof(object->type));
  err |= read_long(socket, &object->id);
  err |= read_char(socket, &object->is_fault_tolerant);
  err |= read_char(socket, &object->debug);
  
  return(err);
}


/* NEW_COMMAND
 *
 * This routine creates a new command structure. Ideally, this routine would
 * pick commands off of a linked list to reduce the overhead of allocating
 * memory so often.
 */
command *new_command()
{
  command *new;

  new = (command *) malloc(sizeof(command));
  new->comm = 0;
  new->message_id = -1;

  return(new);
}


/* _READ_COMMAND
 *
 * This routine is responsible for reading in a command. It is already assumed
 * that the size of the command has been read. This routine is normally called
 * when a higher mechanism is certain that enough bytes are ready to be
 * read. This prevents this routine from blocking while waiting for network
 * data.
 */
int _read_command(c, socket)
     command *c;
     int socket;
{
  int err=0, k;
  short num_machines;
  command *temp_command;

  err |= read_header(socket, &c->comm);      /* What command are we reading */

  err |= read_long(socket, &c->message_id);  /* The message id number */
  err |= read_long(socket, &c->time_sent);   /* When was this message sent */
  if (c->time_sent > current_time)           /* Possibly adjust our time */
    current_time = c->time_sent + 1;
  err |= read_long(socket, &c->recv_ref);    /* Our reference number */
  err |= read_long(socket, &c->send_ref);    /* The sender's reference num */

  c->socket = socket;                        /* What socket read from */
  
  switch(c->comm) {
  case REQUEST_ID:
    err |= read_char(socket, (char *) &c->data.req_id.connect_id);
    break;                                    /* No more data to read */
  case NEW_OBJECT:
    err |= read_object(socket, &c->data.new_o.object);
    break;
  case CREATION_FAILED:
    err |= read_short(socket, &c->data.n_created.error_code);
    break;
  case OBJECT_CREATED:
    break;
  case DELETE_OBJECT:
    break;
  case OBJECT_DELETED:
    break;
  case OBJECT_LOAD:
    err |= read_long(socket, &c->data.obj_load.proc_id);
    err |= read_string(socket, c->data.obj_load.object_name, 32);
    err |= read_long(socket, &c->data.obj_load.port_num);
    break;
  case OBJECT_MESSAGE:
    err |= read_short(socket, &c->data.mess.the_message);
    err |= read_params(socket, &c->data.mess.params);
    err |= read_short(socket, &c->data.mess.wait_return);
    break;
  case RETURN_MESSAGE:
    err |= read_long(socket, &c->data.return_m.result);
    err |= read_params(socket, &c->data.return_m.params);
    break;
  case OBJECT_LOADED:
    err |= read_short(socket, &c->data.done_load.status);
    break;
  case SPEED_LINK:
  case LINK_OBJECT:
    break;
  case OBJECT_LINKED:
    err |= read_short(socket, &c->data.o_linked.status);
    err |= read_object(socket, &c->data.o_linked.object);
    break;
  case SPEED_MADE:
    err |= read_short(socket, &c->data.speed_made.status);
    err |= read_long(socket, &c->data.speed_made.port_num);
    break;
  case REGISTER_SERV:
    err |= read_string(socket, c->data.reg_serv.type, 30);
    err |= read_string(socket, c->data.reg_serv.machine, 33);
    err |= read_long(socket, &c->data.reg_serv.port_num);
    break;
  case LOG_MESSAGE:
    temp_command = new_command();
    err |= read_command(temp_command, socket);
    c->data.log.comm = temp_command;
    break;
  case GET_CLASS_INFO:
    err |= read_short(socket, &c->data.class.control);
    err |= read_short(socket, &c->data.class.num);
    err |= read_string(socket, c->data.class.class, sizeof(c->data.class.class));
    break;
  case R_CLASS_INFO:
    err |= read_short(socket, &num_machines);
    c->data.r_class.num_machines = num_machines;
    c->data.r_class.machine = (char **) malloc(sizeof(char *) * num_machines);
    for(k=0; k < num_machines; k++)
      err |= read_a_string(socket, &c->data.r_class.machine[k]);
    break;
  case LAUNCH_SERVICE:
    err |= read_string(socket, c->data.l_serv.service, 33);
    err |= read_string(socket, c->data.l_serv.machine, 33);
    break;
  case GET_TIME:
    break;
  case SEND_TIMES:
    err |= read_double(socket, &c->data.send_times.time_idle);
    err |= read_double(socket, &c->data.send_times.time_computing);
    err |= read_double(socket, &c->data.send_times.time_writing);
    break;
  case QUERY_SERV:
    err |= read_string(socket, c->data.query_serv.machine, 33);
    err |= read_string(socket, c->data.query_serv.service, 33);
    break;
  case SERV_INFO:
    err |= read_string(socket, c->data.serv_info.machine, 33);
    err |= read_string(socket, c->data.serv_info.service, 33);
    err |= read_long(socket, &c->data.serv_info.port_num);
    err |= read_short(socket, &c->data.serv_info.running);
    break;
  case BASE_DIR:
    err |= read_string(socket, c->data.base_dir.base, 100);
    break;
  case START_TIME:
  case KILL:
  case NO_OP:
    break;
  default:
    break;
  }
  return(err);
}


/* READ_COMMAND
 *
 * This routine will read in the number of bytes coming across for a command
 * and then call _read_command to read in the information.
 */
int read_command(c, socket)
     command *c;
     int socket;
{
  long packet_size;
  int err=0;

  err = read_block(socket, &packet_size, sizeof(long));
  if (err) {
    printf("Error reading first block\n");
    return(err);
  }
  packet_size = ntohl(packet_size);
  err = _read_command(c, socket);
  return(err);
}


/* WRITE_COMMAND
 *
 * This routine is the sister of _read_command. It writes a command to a
 * buffer, which is later flushed out on the net. 
 */
int write_command(c, socket)
     command *c;
     int socket;
{
  int k;
  long packet_size;

  write_header(c->comm);                /* Which command is being sent */
  write_long(c->message_id);            /* What is the message id */
  write_long(current_time);             /* The time it was sent */
  current_time++;                       /* Increment our relative time count */
  write_long(c->recv_ref);              /* The destination's ref num */
  write_long(c->send_ref);              /* Our ref num */

  switch(c->comm) {
  case REQUEST_ID:
    write_char(c->data.req_id.connect_id);
    break;
  case NEW_OBJECT:
    write_object(&c->data.new_o.object);
    break;
  case CREATION_FAILED:
    write_short(c->data.n_created.error_code);
    break;
  case OBJECT_CREATED:
    break;
  case DELETE_OBJECT:
    break;
  case OBJECT_DELETED:
    break;
  case OBJECT_LOAD:
    write_long(c->data.obj_load.proc_id);
    write_string(c->data.obj_load.object_name);
    write_long(c->data.obj_load.port_num);
    break;
  case OBJECT_MESSAGE:
    write_short(c->data.mess.the_message);
    write_params(c->data.mess.params);
    write_short(c->data.mess.wait_return);
    break;
  case RETURN_MESSAGE:
    write_long(c->data.return_m.result);
    write_params(c->data.return_m.params);
    break;
  case OBJECT_LOADED:
    write_short(c->data.done_load.status);
    break;
  case OBJECT_LINKED:
    write_short(c->data.o_linked.status);
    write_object(&c->data.o_linked.object);
    break;
  case LINK_OBJECT:
  case SPEED_LINK:
    break;
  case SPEED_MADE:
    write_short(c->data.speed_made.status);
    write_long(c->data.speed_made.port_num);
    break;
  case REGISTER_SERV:
    write_string(c->data.reg_serv.type);
    write_string(c->data.reg_serv.machine);
    write_long(c->data.reg_serv.port_num);
    break;
  case LOG_MESSAGE:
    write_command(c->data.log.comm, socket);
    break;
  case GET_CLASS_INFO:
    write_short(c->data.class.control);
    write_short(c->data.class.num);
    write_string(c->data.class.class);
    break;
  case R_CLASS_INFO:
    write_short(c->data.r_class.num_machines);
    for(k=0; k < c->data.r_class.num_machines; k++)
      write_string(c->data.r_class.machine[k]);
    break;
  case LAUNCH_SERVICE:
    write_string(c->data.l_serv.service);
    write_string(c->data.l_serv.machine);
    break;
  case GET_TIME:
    break;
  case SEND_TIMES:
    write_double(c->data.send_times.time_idle);
    write_double(c->data.send_times.time_computing);
    write_double(c->data.send_times.time_writing);
    break;
  case QUERY_SERV:
    write_string(c->data.query_serv.machine);
    write_string(c->data.query_serv.service);
    break;
  case SERV_INFO:
    write_string(c->data.serv_info.machine);
    write_string(c->data.serv_info.service);
    write_long(c->data.serv_info.port_num);
    write_short(c->data.serv_info.running);
    break;
  case BASE_DIR:
    write_string(c->data.base_dir.base);
    break;
  case START_TIME:
  case KILL:
  case NO_OP:
    break;
  default:
/*    printf("ERROR!!! WRITE_COMMAND unknown command %d\n", c->comm); */
    break;
  }
  

  packet_size = get_num_bytes();           /* How many bytes are we sending */
  packet_size = htonl(packet_size);        /* Make into network format */
                                           /* Write out command size first */
  _write_data(socket, (char *) &packet_size, sizeof(long));

  return(flush_buff(socket));              /* Write out the command */
}


/* OPEN_COMMANDS
 *
 * This routine is called whenever an application is going to perform buffered
 * reading on a socket. A structure must be allocated to keep track of the
 * buffered reads.
 */
void open_commands(socket)
     int socket;
{

  /* Allocate the structure off of the array 'comm_array' */

  comm_array[socket] = (comm_ent *) malloc(sizeof(comm_ent));
  comm_array[socket]->bytes_needed = 0;
  comm_array[socket]->bytes_read = 0;
  comm_array[socket]->size_of_buff = DEFAULT_BUFF_SIZE;
  comm_array[socket]->data = (char *) malloc(DEFAULT_BUFF_SIZE);
}


/* CLOSE_COMMANDS
 *
 * This routine should be called when no more buffered I/O needs to be
 * performed on a socket
 */
void close_commands(socket)
     int socket;
{
  free(comm_array[socket]->data);
  free(comm_array[socket]);
}


/* COMPLETE_COMMAND
 *
 * This routine should be used to read commands when buffered I/O is desired.
 * The routine will load in bytes from the net and only call _read_command
 * when the entire command has been received. The routine will return 0 if
 * no command was read, 1 if a command was read, or -1 if a network error
 * occured.
 */
int complete_command(comm, socket)
     command *comm;
     int socket;
{
  int err=0;
  long packet_size, bytes_read;
  comm_ent *ce;


#ifdef NEW_IO
  ce = comm_array[socket];
  if (ce->bytes_needed == 0) {          /* Need more bytes to complete? */
    err = read_block(socket, &packet_size, sizeof(long));
    if (err)                            /* Read in a new command        */
      return(-1);
    packet_size = ntohl(packet_size);
    ce->bytes_needed = packet_size;     /* How many bytes in the command */
    if (ce->size_of_buff < ce->bytes_needed) {    /* Grow the buffer */
      ce->data = (char *) realloc(ce->data, ce->bytes_needed);
      ce->size_of_buff = ce->bytes_needed;
    }
    ce->bytes_read = 0;                 /* We have yet to read the command */
  }

  /* Try to read in as many bytes from the command as possible */

  bytes_read = read_chunk(socket, ((char *)ce->data) + ce->bytes_read, 
			  ce->bytes_needed);
  if (bytes_read < 0) {                 /* There was an error reading */
    return(-1);
  }

  ce->bytes_read += bytes_read;       /* Update our bytes read count */
  ce->bytes_needed -= bytes_read;     /* Update our bytes needed count */
  if (ce->bytes_needed == 0) {        /* Do we have the entire command */
    set_read(BUFFER, ce->data);       /* Yes, set the internal read buffer */
    err |= _read_command(comm, socket);  /* Read command from internal buff */
    set_read(SOCKET, 0);              /* Set reading back to sockets */
    return(1);                        /* We have a complete command! */
  }
  return(0);                          /* Still more data to go */

#else
  err = read_command(comm, socket);   /* Do normal I/O which can block */ 
  if (err)
    return(-1);
  else
    return(1);                        /* We are assured of getting a comm */
#endif
}


/* COMPLETE_A_COMMAND
 *
 * This routine is similiar to complete_command except that it will allocate
 * the return command structure.
 */
int complete_a_command(comm, socket)
     command **comm;
     int socket;
{
  int result;
  command n_comm;

  result = complete_command(&n_comm, socket);
  if (result == 1) {
    *comm = new_command();
    *(*comm) = n_comm;
    return(1);
  }
  else
    return(result);
}    



/***************************************************************************
 *                                                                         *
                         Command Building Routines
 *                                                                         *
 ***************************************************************************/


void build_REQUEST_ID(c, connect_id)
     command *c;
     int connect_id;
{
  c->comm = REQUEST_ID;
  c->message_id = 0;
  c->data.req_id.connect_id = connect_id;
}


void build_NEW_OBJECT(c, mess_id, ref, object)
     command *c;
     long mess_id;
     long ref;
     object_instance *object;
{
  c->comm = NEW_OBJECT;
  c->message_id = mess_id;
  c->send_ref = ref;
  c->data.new_o.object = *object;
}
  

void build_OBJECT_CREATED(c, mess_id, ref, s_ref)
     command *c;
     long mess_id;
     long ref;
     long s_ref;
{
  c->comm = OBJECT_CREATED;
  c->message_id = mess_id;
  c->recv_ref = ref;
  c->send_ref = s_ref;
}


void build_LINK_OBJECT(c, ref, s_ref)
     command *c;
     long ref;
     long s_ref;
{
  c->comm = LINK_OBJECT;
  c->recv_ref = ref;
  c->send_ref = s_ref;
}


void build_SPEED_LINK(c, ref)
     command *c;
     long ref;
{
  c->comm = SPEED_LINK;
  c->recv_ref = ref;
}


void build_SPEED_MADE(c, status, ref, port_num)
     command *c;
     int status;
     long ref;
     long port_num;
{
  c->comm = SPEED_MADE;
  c->data.speed_made.status = status;
  c->send_ref = ref;                        /* The object's ref num */
  c->data.speed_made.port_num = port_num;
}


void build_OBJECT_LINKED(c, status, ref, object)
     command *c;
     int status;
     long ref;
     object_instance *object;
{
  c->comm = OBJECT_LINKED;
  c->data.o_linked.status = status;
  c->recv_ref = ref;
  c->data.o_linked.object = *object;
}


void build_CREATION_FAILED(c, mess_id, ref, errcode)
     command *c;
     long mess_id;
     long ref;
     int errcode;
{
  c->comm = CREATION_FAILED;
  c->message_id = mess_id;
  c->data.n_created.error_code = errcode;
  c->recv_ref = ref;
}


void build_DELETE_OBJECT(c, mess_id, ref, r_ref)
     command *c;
     long mess_id;
     long ref;
     long r_ref;
{
  c->comm = DELETE_OBJECT;
  c->message_id = mess_id;
  c->recv_ref = ref;
  c->send_ref = r_ref;
}


void build_OBJECT_DELETED(c, mess_id, ref)
     command *c;
     long mess_id;
     long ref;
{
  c->comm = OBJECT_DELETED;
  c->message_id = mess_id;
  c->recv_ref = ref;
}


void build_OBJECT_LOAD(c, mess_id, obj_name, ref, proc_id, port_num)
     command *c;
     long mess_id;
     char *obj_name;
     long ref;
     int proc_id;
     long port_num;
{
  int len = strlen(obj_name);

  c->comm = OBJECT_LOAD;
  c->message_id = mess_id;
  strncpy(c->data.obj_load.object_name, obj_name, 30);
  c->data.obj_load.object_name[(31 > len) ? len : 31] = 0;
  c->send_ref = ref;
  c->data.obj_load.proc_id = proc_id;
  c->data.obj_load.port_num = port_num;
}


void build_OBJECT_LOADED(c, ref)
     command *c;
     long ref;
{
  c->comm = OBJECT_LOADED;
  c->send_ref = ref;
}


void build_OBJECT_MESSAGE(c, ref, r_ref, mess_id, the_mess, params, wait_r)
     command *c;
     long ref;
     long r_ref;
     long mess_id;
     int the_mess;
     pblock *params;
     int wait_r;
{
  c->comm = OBJECT_MESSAGE;
  c->message_id = mess_id;
  c->data.mess.the_message = the_mess;
  c->data.mess.params = params;
  c->recv_ref = ref;
  c->send_ref = r_ref;
  c->data.mess.wait_return = wait_r;
}


void build_RETURN_MESSAGE(c, ref, mess_id, result, params)
     command *c;
     long ref;
     long mess_id;
     long result;
     pblock *params;
{
  c->comm = RETURN_MESSAGE;
  c->message_id = mess_id;
  c->recv_ref = ref;
  c->data.return_m.result = result;
  c->data.return_m.params = params;
}


void build_REGISTER_SERV(c, type, machine, port_num)
     command *c;
     char *type;
     char *machine;
     long port_num;
{
  c->comm = REGISTER_SERV;
  strcpy(c->data.reg_serv.type, type);
  strcpy(c->data.reg_serv.machine, machine);
  c->data.reg_serv.port_num = port_num;
}


void build_GET_CLASS_INFO(c, messid, ref, control, num, class)
     command *c;
     long messid;
     long ref;
     int control;
     int num;
     char *class;
{
  c->comm = GET_CLASS_INFO;
  c->message_id = messid;
  c->send_ref = ref;
  c->data.class.control = control;
  c->data.class.num = num;
  strcpy(c->data.class.class, class);
}


void build_R_CLASS_INFO(c, messid, ref, num, machines)
     command *c;
     long messid;
     long ref;
     int num;
     char **machines;
{
  c->comm = R_CLASS_INFO;
  c->message_id = messid;
  c->recv_ref = ref;
  c->data.r_class.num_machines = num;
  c->data.r_class.machine = machines;
}


void build_NO_OP(c)
     command *c;
{
  c->comm = NO_OP;
}


void build_KILL(c)
     command *c;
{
  c->comm = KILL;
}


void build_LAUNCH_SERVICE(c, mess_id, service, machine)
     command *c;
     int mess_id;
     char *service;
     char *machine;
{
  c->comm = LAUNCH_SERVICE;
  c->message_id = mess_id;
  strcpy(c->data.l_serv.service, service);
  strcpy(c->data.l_serv.machine, machine);
}


void build_START_TIME(c, ref)
     command *c;
     long ref;
{
  c->comm = START_TIME;
  c->recv_ref = ref;
}


void build_GET_TIME(c, messid, recv_ref, send_ref)
     command *c;
     long messid;
     long recv_ref;
     long send_ref;
{
  c->comm = GET_TIME;
  c->message_id = messid;
  c->recv_ref = recv_ref;
  c->send_ref = send_ref;
}


void build_SEND_TIMES(c, messid, ref, time_idle, time_computing, time_writing)
     command *c;
     long messid;
     long ref;
     double time_idle, time_computing, time_writing;
{
  c->comm = SEND_TIMES;
  c->message_id = messid;
  c->data.send_times.time_idle = time_idle;
  c->data.send_times.time_computing = time_computing;
  c->data.send_times.time_writing = time_writing;
  c->recv_ref = ref;
}


void build_QUERY_SERV(c, messid, machine, service)
     command *c;
     long messid;
     char *machine;
     char *service;
{
  c->comm = QUERY_SERV;
  c->message_id = messid;
  strcpy(c->data.query_serv.machine, machine);
  strcpy(c->data.query_serv.service, service);
}


void build_SERV_INFO(c, messid, machine, service, port_num, running)
     command *c;
     long messid;
     char *machine;
     char *service;
     long port_num;
     int running;
{
  c->comm = SERV_INFO;
  c->message_id = messid;
  strcpy(c->data.serv_info.machine, machine);
  strcpy(c->data.serv_info.service, service);
  c->data.serv_info.port_num = port_num;
  c->data.serv_info.running = running;
}


void build_BASE_DIR(c, dir)
     command *c;
     char *dir;
{
  c->comm = BASE_DIR;
  strcpy(c->data.base_dir.base, dir);
}



/* CLEAN_COMMAND
 *
 * This routine will deallocate any external memory associated with a
 * command, such as a parameter_block.
 */
void clean_command(c)
     command *c;
{
  int k;

  switch(c->comm) {
  case OBJECT_MESSAGE:
    free_params(c->data.mess.params);
    break;
  case RETURN_MESSAGE:
    free_params(c->data.return_m.params);
    break;
  case LOG_MESSAGE:
    free_command(c->data.log.comm);
    break;
  case R_CLASS_INFO:
    for(k=0; k < c->data.r_class.num_machines; k++)
      free(c->data.r_class.machine[k]);
    free(c->data.r_class.machine);
    break;
  }
  c->comm = 0;
}


/* FREE_COMMAND
 *
 * Free the command by first cleaning it and then deallocating command
 * memory.
 */
void free_command(c)
     command *c;
{
  clean_command(c);
  free(c);
}
