/* timed_task.c
 * Jim Gochee
 * 3/9/92
 *
 * This file contains routines to handle lists of tasks that need to be
 * done at specific intervals. The granularity of these tasks is currently
 * 1 second. The timedtask functions are used as follows:
 *
 * Create a timedtask structure with 'new_timed_tasks'.
 * Then call 'add_timed_task' for each task you wish to install. With each
 *   task is given the function to callback and a parameter. 'Add_timed_task'
 *   returns an id number which can be used to remove the task at a later time.
 * To find out how much time can pass before the next task needs to execute,
 *   call 'max_task_time'.
 * To allow tasks to execute, call 'do_tasks'.
 * To remove all timed tasks and free memory, call 'free_timed_tasks'
 */
#include <sys/time.h>

#include "List.h"
#include "timed_task.h"

/*
 * This is what each task has associated with it
 */
typedef struct _task_entry {
  struct timeval delay;         /* How long to wait between running the task */
  struct timeval run_time;      /* When this task should run next (in secs)  */
  int id;                       /* An id for this task. Used to later delete */
  int (*f)();                   /* The function to be called to do the task  */
  void *data;                   /* A parameter to be sent to the function    */
} task_entry;


static TASK_ID = 1;             /* The 'id' of each task is based on this */


int timecmp(t1, t2)
     struct timeval *t1, *t2;
{
  if (t1->tv_sec < t2->tv_sec || ((t1->tv_sec == t2->tv_sec) && t1->tv_usec < t2->tv_usec))
    return(-1);
  
  if (t1->tv_sec == t2->tv_sec && t1->tv_usec == t2->tv_usec)
    return(0);
  
  return(1);
}



/* NEW_TIMED_TASKS
 *
 * This routine allocates the necessary 'List', which is used to keep track
 * of the tasks.
 */
timedtasks new_timed_tasks()
{
  List l;

  l = CreateList();
  return((timedtasks) l);
}


/* _INSERT_TIMED_ENTRY
 *
 * This routine will take a given timed entry and insert it into the 'to do'
 * list based on when the task will run. The list is in sorted order by the
 * run times.
 */
void _insert_timed_entry(ts, te)
     timedtasks ts;
     task_entry *te;
{
  List l;
  task_entry *temp_te;

  l = (List) ts;
  SetBeginning(l);
  while(!EndList(l)) {
    ReadData(l, 0, (void **) &temp_te);
    if (timecmp(&te->run_time, &temp_te->run_time) <= 0)
      break;
    SetNext(l);
  }
  
  InsertData(l, 0, (void *) te);
}


/* ADD_TIMED_TASK
 *
 * This routine creates a new timed task and inserts it into the list of timed
 * tasks. 
 */
long add_timed_task(ts, delay_secs, delay_usecs, f, data)
     timedtasks ts;
     long delay_secs;
     long delay_usecs;
     void *f;
     void *data;
{
  task_entry *te;
  struct timezone tz;

  te = (task_entry *) malloc(sizeof(task_entry));
  te->delay.tv_sec = delay_secs;
  te->delay.tv_usec = delay_usecs;
  te->f = f;
  te->data = data;
  te->id = TASK_ID++;
  
  gettimeofday(&te->run_time, &tz);
  te->run_time.tv_sec += te->delay.tv_sec;
  te->run_time.tv_usec += te->delay.tv_usec;
  _insert_timed_entry(ts, te);
  return(te->id);
}


/* REMOVE_TIMED_TASK
 *
 * This routine will remove a task described by an 'id'.
 */
void remove_timed_task(ts, id)
     timedtasks ts;
     long id;
{
  List l;
  task_entry *te;
 
  l = (List) ts;
  SetBeginning(l);
  while(!EndList(l)) {
    ReadData(l, 0, (void **) &te);
    if (te->id == id) {
      free(te);
      DeleteNode(l);
      return;
    }
    SetNext(l);
  }
}


/* FREE_TIMED_TASKS
 *
 * This routine deallocates all memory used by the timed task.
 */
void free_timed_tasks(ts)
     timedtasks ts;
{
  List l;
  task_entry *te;

  l = (List) ts;

  SetBeginning(l);
  while(!EndList(l)) {
    ReadData(l, 0, (void **) &te);
    free(te);
    DeleteNode(l);
  }
  FreeList(l);
}


/* MAX_TASK_TIME
 *
 * This function returns the number of seconds until the next task in the
 * queue needs to be run.
 */
void next_task_wait(ts, tv)
     timedtasks ts;
     struct timeval *tv;
{
  List l;
  task_entry *te;
  struct timeval cur;
  struct timezone tz;

  l = (List) ts;
  SetBeginning(l);
  if (EndList(l)) {                         /* There are no tasks to run */
    tv->tv_sec = 30000;
    tv->tv_usec= 0;
    return;
  }

  ReadData(l, 0, (void **) &te);
  gettimeofday(&cur, &tz);
  tv->tv_sec = te->run_time.tv_sec - cur.tv_sec;
  tv->tv_usec = te->run_time.tv_usec - cur.tv_usec;


  if (tv->tv_usec < 0) {                   /* Do we need to borrow a second  */
    tv->tv_sec--;
    tv->tv_usec += 1000000;                /* One second is worth 1000 msecs */
  }

  if (tv->tv_sec < 0) {                    /* Did we miss the task?          */
    tv->tv_sec = 0;
    tv->tv_usec = 0;
  }
}


/* DO_TASKS
 *
 * This routine should be called periodically to run timed tasks. If a timed
 * tasks returns a value other than 0, then do_tasks will return immediately
 * with that result, otherwise several tasks can proceed
 */
int do_tasks(ts)
     timedtasks ts;
{
  List l;
  int result;
  task_entry *te;
  struct timeval cur;
  struct timezone tz;

  gettimeofday(&cur, &tz);

  l = (List) ts;
  SetBeginning(l);

  while(!EndList(l)) {
    ReadData(l, 0, (void **) &te);
    if (timecmp(&te->run_time, &cur) > 0)       /* No more left to run    */
      return(0);                                /* Returning '0' means OK */
    DeleteNode(l);
    te->run_time.tv_sec = cur.tv_sec + te->delay.tv_sec;
    te->run_time.tv_usec = cur.tv_usec + te->delay.tv_usec;

    _insert_timed_entry(ts, te);                /* Put back in at new time */

    result = (*te->f)(te->data);

    if (result)                                 /* Was there an error?     */
      return(result);                           /* Return error to caller  */
  }
  return(0);
}

	
