/* MAND
 * Jim Gochee
 * 5/10/92
 *
 * This object, when used with the PDPE (Personal Distributed Programming 
 * Environment) will draw a mandelbrot image utilizing multiple cpus.
 */
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>

/* Include files necessary for objects */
#include "sockets.h"
#include "spede.h"
#include "spede_io.h"
#include "objects.h"
#include "objlib.h"

/* Non-necessary include files */
#include "misc.h"

#define TRUE 1
#define FALSE  0
#define MAX_OBJECTS 100
#define ZOOM_FACTOR 2.5             /* The closer to 2, the bigger the zoom */

#define LINE_INC 7

#define SET_JOB     10              /* Object messages */
#define START_TASK  11

/**
 ** Some typedefs describing the image and what a 'job' is.
 **/
struct mandel_task {
  int width, height;
  double left, right, top, bottom;
  double cr, ci;
};

struct mandel_job2 {
  int top, bottom;
};

struct mandel_job {
  int left, right, top, bottom;
  int width, height;
  unsigned char *pixels;
  int x,y;
};



char *object_name = "mand";                /* This name is the object name */
object_instance *task_obj[MAX_OBJECTS];    /* An array of object pointers  */
unsigned char *final_pix;                  /* Pointer to the image in memory */
struct mandel_task g_task;                 /* The mandelbrot descriptor    */
int end_line, current_line, num_working;   /* Used to monitor image lines  */
pblock *params;                            /* Params for work objects      */
int num_machines;                          /* The num of available nodes   */
int num_objects;                           /* The num of work objects in use */
char class_name[256];                      /* Which classes to use         */
int exact_mode, line_increment;
int use_local = FALSE;
double setup_time = 0.0;

extern char *optarg;

/* LOCAL Mandelbrot functions
 *
 */

int mandel_do_pix(task, job)
     struct mandel_task *task;
     struct mandel_job *job;
{
  double rx, ry;

  rx  = task->left + ((task->right - task->left) * (double) job->x) / (double) task->width;
  ry = task->top + ((task->bottom - task->top) * (double) job->y) / (double) task->height;

  job->pixels[(job->y - job->top)*job->width + job->x - job->left] = mandel(rx, ry, task->cr, task->ci, 256, 1);
  if (++job->x == job->right) {
    job->x = job->left;
    job->y++;
    if (job->y > job->bottom)
      return 1;
  }
  return 0;
}

int
mandel_do_job(task, job)
     struct mandel_task *task;
     struct mandel_job *job;
{
  do
    if (mandel_do_pix(task, job))
      return 1;
  while (TRUE);
  return 0;
}

/* mandel.c: mandelbrot routine
 * From Fractint public domain fractal-manipulation program.
 */
int
mandel(initreal,initimag,parm1,parm2,maxit,inside)
double initreal,initimag,parm1,parm2;
int maxit,inside;
{
   double oldreal, oldimag, newreal, newimag, magnitude;
   int color;

   oldreal=parm1;
   oldimag=parm2;
   magnitude = 0.0;
   color = 0;
   while ((magnitude < 4.0) && (color < maxit)) {
      newreal = oldreal * oldreal - oldimag * oldimag + initreal;
      newimag = 2 * oldreal * oldimag + initimag;
      color++;
      oldreal = newreal;
      oldimag = newimag;
      magnitude = newreal * newreal + newimag * newimag;
   }
   if (color >= maxit) color = inside;
   return(color);
}
  


/* CALLBACK
 *
 * This routine is the callback for when an object receives a OBJ_MESSAGE
 * command. This object will never receive one of those, so this routine is
 * not used.
 */
static long callback(message, params, r_params)
     int message;
     parameter_block *params, **r_params;
{
  long result;
  parameter_block param;

  param.num_params = 0;
  if (r_params != 0)
    *r_params = &param;

  switch(message) {
  case CREATE:
    break;
  case DELETE:
    break;
  default:
    printf("Invalid message.\n");
    return(0);
  }
}
  

/* MAIN
 *
 * This is the same for every object.
 */
main(argc, argv) 
     int argc;
     char **argv;
{
  setup_time = get_seconds();
  if(init_objects(argc, argv, callback)) {
    perror("Error in init_objects");
    exit(1);
  }

  register_dean();                      /* Tell local dean we exist */
}



/* INIT_TASK
 *
 * This routine sets up the mandelbrot image to be computed.
 */
void init_task(task)
     struct mandel_task *task;
{
  long size;

  task->width = 500;
  task->height = 370;

  task->left = -2.5; 
  task->right = 1.5;
  task->top = -1.5;
  task->bottom = 1.5;

/*  task->left = -0.6633725;
  task->right = -0.6533721;
  task->top = -0.45337413;
  task->bottom = -0.4458738;
*/

  task->cr = 0;
  task->ci = 0;

  size = task->width * task->height * sizeof(unsigned char);
  final_pix = (unsigned char *) malloc(size);
}


/* CREATE_OBJECTS
 *
 * This routine is responsible for creating the work objects to compute the
 * actual image. The machines that will be used are gotten from the command
 * line in the form of a class name. The 'reg' daemon is queried to find out
 * which machines are available in the given class. Note that machine names
 * could also have been entered on the command line with or in place of a
 * class name. Using a class name only is a choice made by the object and not
 * the restriction of the underlying mechanism.
 */
void create_objects()
{
  char *machine;
  int k, m, err;
  object_instance *obj;
  link_des link;
  char **the_machines;
  extern object_descriptor run_info;

  printf("Creating the work objects\n");

  /* Query the 'reg' daemon for machines of the given class */
  num_machines = run_info.num_machines;

  num_objects = (run_info.num_objects > MAX_OBJECTS) ? MAX_OBJECTS : 
    run_info.num_objects;
  
  printf("Creating from a list of %d machines\n", num_machines, class_name);

  printf("Attempting to creating %d objects\n", num_objects);
  m = 0;
  for(k=0; k < num_objects;) {
    if (num_machines < 1) {
      printf("There are no machines specified!!\n");
      exit(0);
    }
    machine = run_info.machine[m % num_machines];
    printf("Asynchronously creating object %d on machine '%s'\n", k, machine);
    task_obj[k] = create_object(machine, "mand_comp", FALSE, FALSE, TRUE);
    if (task_obj[k] != 0) {
      k++;
    }
    else {
      printf("Failed creating on that machine...trying another.\n");
      run_info.machine[m] = run_info.machine[num_machines-1];
      num_machines--;
    }
    m++;
  }
  num_objects = k;

  if (exact_mode)
    line_increment = (g_task.height+num_objects-1)/num_objects;

  if (num_objects < 1) {
    printf("There are no objects\n");
    exit(0);
  }

  err = sync_created_objects();                /* Wait until all are created */
  if (err) {
    printf("There was an error creating an object!!!\n");
  }

  printf("%d objects were actually created\n", num_objects);
  
#ifdef SPEEDLINKIT
  for(k=0; k < num_objects; k++) {
    /* These next lines form a speed link, which bypasses the dean */
    link = task_obj[k]->location;             /* Link to the object location */
    task_obj[k] = speed_link(&link);          /* Form the link               */
  }
#endif
  printf("Setup cost was %.4g seconds\n", get_seconds()-setup_time);

}


void reset_work(task)
     struct mandel_task *task;
{
  current_line = 0;
  end_line = task->height - 1;
}


int more_work(start, stop)
     int *start, *stop;
{
  if (current_line >= end_line)
    return(FALSE);

  *start = current_line;
  *stop = current_line + line_increment;
  if (*stop > end_line)
    *stop = end_line;

  current_line = *stop+1;
  return(TRUE);
}
  


/* OBJECT_DONE
 *
 * This routine is called by the object library when a return message has
 * arrived.
 */
void object_done(object, rinfo, obj_num)
     object_instance *object;
     return_info *rinfo;
     long obj_num;
{
  int *p_start, start, stop, pix_start;
  struct mandel_task *task = &g_task;
  void *data;
  long len;

  get_params(rinfo->r_params, &p_start, &data);
  get_param_len(rinfo->r_params, 0, &len);

  pix_start = start = ntohl(*p_start);
  memcpy(final_pix + start*task->width, data, len);

/*  printf("Object %d is done with %d to %d\n", obj_num, start,
	 start+len/task->width); */

  /* See if there is more work to do */
  if (more_work(&start, &stop)) {
    object_work(obj_num, object, params, start, stop);
  }
  else {
    num_working--;                  /* One less object is working */
  }
  draw_pixmap(final_pix + pix_start*task->width, 0, pix_start, g_task.width, len/task->width);

}


/* OBJECT_WORK
 *
 * This routine is called when some work is being sent to an object.
 */
object_work(k, object, params, start, stop)
     int k;
     object_instance *object;
     pblock *params;
     int start, stop;
{
  int *start_p, *stop_p, *obj_num;

  /* Get pointers to the param data */
  get_params(params, &obj_num, 0, &start_p, &stop_p);

  *obj_num = k;                           /* Object number */
  *start_p = htonl(start);                /* Starting line */
  *stop_p = htonl(stop);                  /* Ending line */

/*  printf("Assigning object %d image line %d to %d\n", k, start, stop);  */

  /* Send this message with no return expected */
  send_obj_message(object, SET_JOB, params, TRUE, FALSE);
  
  /* Send the command to start work, with a callback when the return */
  /* message arrives                                                 */
  send_obj_mess_callback(object, START_TASK, params, object_done, (void *) k);
}


/* MANDEL_DO_TASK
 *
 * This routine is called to start the computation of the image.
 */
void mandel_do_task(task)
     struct mandel_task *task;
{
  char task_string[256];
  struct mandel_job2 job;
  int k, start, stop, *start_p, *stop_p, *obj_num, num_completed;
  double start_time;
  double idle, cpu, writing;

  start_time = get_seconds();            /* Start timing the draw */

  /* Build the task descriptor, which happens to be an ascii string */
  sprintf(task_string, "%d %d %.18g %.18g %.18g %.18g %.18g %.18g",
	  task->width, task->height, task->left, task->right, task->top,
	  task->bottom, task->cr, task->ci);

  params = new_params(4, sizeof(int), strlen(task_string)+1, sizeof(int), sizeof(int));
  set_params(params, 0, task_string, 0, 0);
  
  reset_work(task);

  num_working=0;                         /* The number of objects working */
  for(k=0; k < num_objects; k++) {
    if (!more_work(&start, &stop)) {
      printf("Warning...no work for object % to do.\n", k);
      continue;
    }
    
    reset_times(task_obj[k]);            /* Reset internal object timer */
    object_work(k, task_obj[k], params, start, stop);
    num_working++;
  }
  
  while(num_working) {                   /* Work until all objects done */
    idle_objects(20, 0);
  }
  
  printf("Time taken to draw pict was %.5g seconds\n", get_seconds() - start_time);


  for(k=0; k < num_objects; k++) {
    get_object_times(task_obj[k], &idle, &cpu, &writing);  /* Get internal timing info */
    printf("Object %d: %.4g secs idle, %.4g secs writing, and %.4g secs computing\n", k, idle, writing, cpu);
  }
  

  /* Don't actually delete objects, since another image might be generated */

/*

  printf("All objects have completed.\n");
  printf("Deleting all objects\n");
  
  for(k=0; k < num_objects; k++)
    delete_object(task_obj[k]);
*/
}


/* DO_LOCAL
 *
 * This routine computes the image locally
 */
void do_local(task)
     struct mandel_task *task;
{
  int start, stop;
  struct mandel_job job;
  double start_time;

  if (exact_mode)
    line_increment = task->height;
  printf("Doing the work locally\n");
  job.left = 0;
  job.right = job.width = task->width;
  job.x = 0;

  start_time = get_seconds();
  reset_work(task);
  while(more_work(&start, &stop)) {
    job.top = start;
    job.bottom = stop;
    job.height = (stop - start + 1);
    job.y = start;
    job.pixels = final_pix + start*task->width;
    mandel_do_job(task, &job);
    draw_pixmap(job.pixels, 0, start, task->width, stop-start+1);
  }
  printf("Total time to draw locally: %.4g seconds\n", get_seconds()-start_time);
}    
  

/* ZOOM_TASK
 * 
 * This routine zooms in on a portion of the image.
 */
void zoom_task(x, y, task, direction)
     int x, y;
     struct mandel_task *task;
     int direction;
{
  float width, height;
  float x_rat, y_rat, difx, dify;

  width = task->right - task->left;
  height = task->bottom - task->top;

  x_rat = width / (float) task->width;
  y_rat = height / (float) task->height;  
  
  difx = (x - (task->width/2)) * x_rat;
  dify = (y - (task->height/2)) * y_rat;

  task->left += difx + (width/ZOOM_FACTOR)*direction;
  task->right += difx - (width/ZOOM_FACTOR)*direction;
  task->top += dify + (height/ZOOM_FACTOR)*direction;
  task->bottom += dify - (height/ZOOM_FACTOR)*direction;
}


void set_options(argc, argv)
     int argc;
     char **argv;
{
  int ch;
  
  exact_mode = FALSE;
  line_increment = LINE_INC;

  while((ch = getopt(argc, argv, "oxl:")) != EOF)
    switch(ch) {
    case 'x':
      exact_mode = TRUE;
      break;
    case 'l':
      line_increment = atoi(optarg);
      break;
    case 'o':
      use_local = TRUE;
      break;
    }
}


/* FREE_LAUNCH
 *
 * This routine is called if the image was launched from the command line.
 */
void free_launch(argc, argv)
     int argc;
     char **argv;
{

  FILE *fp;
  int red[256], green[256], blue[256];
  int i, x, y, button;
  char line[256];
  extern object_descriptor run_info;

  set_options(argc, argv);

  fp = fopen("mapfile", "r");
  if (fp == 0) {
    perror("Could not find mapfile...\n");
    exit(1);
  }
       
  for(i=0; i < 256; i++) {
    if (!fgets(line, 255, fp)) {
      fclose(fp);
      printf("err");
      exit(1);
    }
    sscanf(line, "%d %d %d", red+i, green+i, blue+i);
  }
  fclose(fp);

  init_task(&g_task);
  setup_window(argc, argv, red, green, blue, g_task.width, g_task.height);

  if (use_local) {
    do_local(&g_task);
  }
  else {
    create_objects();
    mandel_do_task(&g_task);
  }

  while(1) {
    go_picture(final_pix, g_task.width, g_task.height, &x, &y, &button);
    zoom_task(x,y, &g_task, button);
    if (use_local)
      do_local(&g_task);
    else
      mandel_do_task(&g_task);
  }
}


/* NORMAL_LAUNCH
 *
 * This routine is entered when an object is created from another object.
 * Mand does not allow this, so exit if it happens.
 */
void normal_launch()
{
  printf("ERROR!!! This object must be launched from the command line\n");
  exit(0);
}
