/* MAND_COMP
 * Jim Gochee
 * 5/10/92
 *
 * This object is a mandelbrot computation object. It is accessed from a 'mand'
 * object. Mand_comp will allow 'mand' to assign it part of an image to
 * compute.
 */
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>

/* Include files necessary for objects */
#include "sockets.h"
#include "spede.h"
#include "spede_io.h"
#include "objects.h"
#include "objlib.h"

#define TRUE 1
#define FALSE  0
#define SET_JOB     10
#define START_TASK  11

/** 
 ** Mandelbrot types
 **/
struct mandel_task {
  int width, height;
  double left, right, top, bottom;
  double cr, ci;
};

struct mandel_job {
  int left, right, top, bottom;
  int width, height;
  unsigned char *pixels;
  int x,y;
};


/**
 ** Globals
 **/
char *object_name = "mand_comp";          /* What is the name of the object */
struct mandel_job *job;
struct mandel_task our_task;
int obj_num;                              /* Our object number */



/**
 ** Mandelbrot computation functions
 **/

int mandel_do_pix(task, job)
     struct mandel_task *task;
     struct mandel_job *job;
{
  double rx, ry;

  rx  = task->left + ((task->right - task->left) * (double) job->x) / (double) task->width;
  ry = task->top + ((task->bottom - task->top) * (double) job->y) / (double) task->height;

  job->pixels[(job->y - job->top)*job->width + job->x - job->left] = mandel(rx, ry, task->cr, task->ci, 256, 1);
  if (++job->x == job->right) {
    job->x = job->left;
    job->y++;
    if (job->y > job->bottom)
      return 1;
  }
  return 0;
}

int
mandel_do_job(task, job)
     struct mandel_task *task;
     struct mandel_job *job;
{
  do
    if (mandel_do_pix(task, job))
      return 1;
  while (TRUE);
  return 0;
}

/* mandel.c: mandelbrot routine
 * From Fractint public domain fractal-manipulation program.
 */
int
mandel(initreal,initimag,parm1,parm2,maxit,inside)
double initreal,initimag,parm1,parm2;
int maxit,inside;
{
   double oldreal, oldimag, newreal, newimag, magnitude;
   int color;

   oldreal=parm1;
   oldimag=parm2;
   magnitude = 0.0;
   color = 0;
   while ((magnitude < 4.0) && (color < maxit)) {
      newreal = oldreal * oldreal - oldimag * oldimag + initreal;
      newimag = 2 * oldreal * oldimag + initimag;
      color++;
      oldreal = newreal;
      oldimag = newimag;
      magnitude = newreal * newreal + newimag * newimag;
   }
   if (color >= maxit) color = inside;
   return(color);
}
  


/* SET_JOB
 *
 * This routine is called when a 'set_job' message is received.
 */
void set_job(params)
     pblock *params;
{
  char *task;
  int *start, *stop;
  struct mandel_task *t = &our_task;
  int *t_obj_num;
  
  get_params(params, &t_obj_num, &task, &start, &stop);
  *start = ntohl(*start);
  *stop = ntohl(*stop);
  obj_num = *t_obj_num;

  /* Get the task description out of the string */
  sscanf(task, "%d %d %lg %lg %lg %lg %lg %lg", &t->width,
	 &t->height, &t->left, &t->right, &t->top,
	 &t->bottom, &t->cr, &t->ci);

  job->left = 0;
  job->right = job->width = t->width;
  job->top = *start;
  job->bottom = *stop;
  job->height = (*stop - *start + 1);
  job->x = 0;
  job->y = *start;
  job->pixels = (unsigned char *) malloc(job->width * job->height * sizeof(unsigned char));

  printf("(%s %d) Will compute from %d to %d\n", object_name, obj_num, *start, *stop);
}


/* DO_WORK
 *
 * This routine is called when the a 'start_task' message has been received.
 */
void do_work(params, r_params)
     pblock *params, **r_params;
{
  pblock *results;
  long k;

  mandel_do_job(&our_task, job);               /* Compute the image section */

  /* Allocate return parameters */
  results = new_params(2, sizeof(long), job->width * job->height);
  k = htonl(job->top);
  set_params(results, &k, job->pixels);
  *r_params = results;  
}



static long callback(message, params, r_params)
     int message;
     parameter_block *params, **r_params;
{
  long result;
  parameter_block param;
  char rw_des[256];

  param.num_params = 0;
  if (r_params != 0)
    *r_params = &param;

  switch(message) {
  case CREATE:
    job = (struct mandel_job *) malloc(sizeof(struct mandel_job));
    break;
  case DELETE:
    break;
  case SET_JOB:
    set_job(params);
    sprintf(rw_des, "%s%d", object_name, obj_num); 
    break;
  case START_TASK:
    printf("Start computation.\n");
    do_work(params, r_params);
    printf("Done with computation.\n");
    break;
  default:
    printf("Invalid message.\n");
    return(0);
  }
}
  

main(argc, argv) 
     int argc;
     char **argv;
{
  if(init_objects(argc, argv, callback)) {
    perror("Error in init_objects");
    exit(1);
  }

  register_dean();                      /* Tell local dean we exist */
}


void free_launch(argc, argv)
     int argc;
     char **argv;
{
  /* Free launch comes here */
  printf("ERROR!!! This object cannot be free-launched\n");
  exit(0);
}

void normal_launch()
{
  /* Nothing to do */
}
